from typing import (
    Optional,
    Sequence,
)
from unittest import (
    TestCase,
    mock,
)

from pcs.cli.common.errors import CmdLineInputError
from pcs.cli.common.parse_args import Argv
from pcs.cli.query import resource
from pcs.common.const import PCMK_STATUS_ROLE_STARTED
from pcs.common.resource_status import (
    MoreChildrenQuantifierType,
    ResourceState,
    ResourceType,
    can_be_promotable,
    can_be_unique,
)
from pcs.common.status_dto import (
    BundleReplicaStatusDto,
    BundleStatusDto,
    CloneStatusDto,
    GroupStatusDto,
    PrimitiveStatusDto,
    ResourcesStatusDto,
)

from pcs_test.tools.misc import dict_to_modifiers


def fixture_primitive_dto(
    resource_id: str,
    instance_id: Optional[str],
    resource_agent: str = "ocf:pacemaker:Dummy",
    node_names: Sequence[str] = ("node1",),
) -> PrimitiveStatusDto:
    return PrimitiveStatusDto(
        resource_id,
        instance_id,
        resource_agent,
        role=PCMK_STATUS_ROLE_STARTED,
        target_role=None,
        active=True,
        orphaned=False,
        blocked=False,
        maintenance=False,
        description=None,
        failed=False,
        managed=True,
        failure_ignored=False,
        node_names=list(node_names),
        pending=None,
        locked_to=None,
    )


def fixture_group_dto(
    resource_id: str,
    instance_id: Optional[str],
    members: list[PrimitiveStatusDto],
) -> GroupStatusDto:
    return GroupStatusDto(
        resource_id, instance_id, False, None, True, False, members
    )


class QueryBaseMixin:
    def setUp(self):
        self.lib = mock.Mock(spec_set=["status"])
        self.lib.status = mock.Mock(spec_set=["resources_status"])
        self.lib_command: mock.Mock = self.lib.status.resources_status

    def test_no_argv(self, mock_print: mock.Mock):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd([])
        self.assertIsNone(cm.exception.message)
        self.lib_command.assert_not_called()
        mock_print.assert_not_called()

    def nonexistent_should_fail(
        self, mock_print: mock.Mock, additional_args: Argv
    ):
        self.lib_command.return_value = ResourcesStatusDto([])
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd(["nonexistent"] + additional_args)
        self.assertEqual(
            cm.exception.message, "Resource 'nonexistent' does not exist"
        )
        self.lib_command.assert_called_once_with()
        mock_print.assert_not_called()


@mock.patch("pcs.cli.query.resource.print")
class TestQueryResourceExists(QueryBaseMixin, TestCase):
    def _call_cmd(self, argv, modifiers=None) -> None:
        modifiers = modifiers or {}
        resource.exists(self.lib, argv, dict_to_modifiers(modifiers))

    def test_too_many_args(self, mock_print):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd(["foo", "bar"])
        self.assertIsNone(cm.exception.message)
        self.lib_command.assert_not_called()
        mock_print.assert_not_called()

    def test_true(self, mock_print):
        self.lib_command.return_value = ResourcesStatusDto(
            [fixture_primitive_dto("primitive", None)]
        )
        with self.assertRaises(SystemExit) as cm:
            self._call_cmd(["primitive"])
        self.assertEqual(cm.exception.code, 0)
        self.lib_command.assert_called_once_with()
        mock_print.assert_called_once_with(True)

    def test_false(self, mock_print):
        self.lib_command.return_value = ResourcesStatusDto(
            [fixture_primitive_dto("primitive", None)]
        )
        with self.assertRaises(SystemExit) as cm:
            self._call_cmd(["nonexistent"])
        self.assertEqual(cm.exception.code, 2)
        self.lib_command.assert_called_once_with()
        mock_print.assert_called_once_with(False)

    def test_quiet(self, mock_print: mock.Mock):
        self.lib_command.return_value = ResourcesStatusDto(
            [fixture_primitive_dto("primitive", None)]
        )
        with self.assertRaises(SystemExit) as cm:
            self._call_cmd(["primitive"], {"quiet": True})
        self.assertEqual(cm.exception.code, 0)
        self.lib_command.assert_called_once_with()
        mock_print.assert_not_called()


@mock.patch("pcs.cli.query.resource.print")
class TestQueryIsType(QueryBaseMixin, TestCase):
    def _call_cmd(self, argv, modifiers=None) -> None:
        modifiers = modifiers or {}
        resource.is_type(self.lib, argv, dict_to_modifiers(modifiers))

    def test_too_many_args(self, mock_print):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd(["foo", "primitive", "abc"])
        self.assertIsNone(cm.exception.message)
        self.lib_command.assert_not_called()
        mock_print.assert_not_called()

    def test_no_type(self, mock_print: mock.Mock):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd(["resource_id"])
        self.assertIsNone(cm.exception.message)
        self.lib_command.assert_not_called()
        mock_print.assert_not_called()

    def test_bad_type(self, mock_print: mock.Mock):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd(["resource_id", "primit"])
        self.assertEqual(
            cm.exception.message,
            (
                "'primit' is not a valid resource type value, use 'bundle', "
                "'clone', 'group', 'primitive'"
            ),
        )
        self.lib_command.assert_not_called()
        mock_print.assert_not_called()

    def test_true(self, mock_print: mock.Mock):
        self.lib_command.return_value = ResourcesStatusDto(
            [fixture_primitive_dto("resource", None)]
        )
        with self.assertRaises(SystemExit) as cm:
            self._call_cmd(["resource", "primitive"])
        self.assertEqual(cm.exception.code, 0)
        self.lib_command.assert_called_once_with()
        mock_print.assert_called_once_with(True)

    def test_false(self, mock_print: mock.Mock):
        self.lib_command.return_value = ResourcesStatusDto(
            [fixture_primitive_dto("resource", None)]
        )
        with self.assertRaises(SystemExit) as cm:
            self._call_cmd(["resource", "group"])
        self.assertEqual(cm.exception.code, 2)
        self.lib_command.assert_called_once_with()
        mock_print.assert_called_once_with(False)

    def test_nonexistent(self, mock_print: mock.Mock):
        self.nonexistent_should_fail(mock_print, ["primitive"])

    def test_promotable_bad_type(self, mock_print: mock.Mock):
        for resource_type in ["primitive", "group", "bundle"]:
            with self.subTest(value=resource_type):
                with self.assertRaises(CmdLineInputError) as cm:
                    self._call_cmd(["resource", resource_type, "promotable"])
                self.assertEqual(
                    cm.exception.message,
                    f"type '{resource_type}' cannot be promotable",
                )
                self.lib_command.assert_not_called()
                mock_print.assert_not_called()

    def test_unique_bad_type(self, mock_print: mock.Mock):
        for resource_type in ["primitive", "group"]:
            with self.subTest(value=resource_type):
                with self.assertRaises(CmdLineInputError) as cm:
                    self._call_cmd(["resource", resource_type, "unique"])
                self.assertEqual(
                    cm.exception.message,
                    f"type '{resource_type}' cannot be unique",
                )
                self.lib_command.assert_not_called()
                mock_print.assert_not_called()

    @mock.patch("pcs.common.resource_status.ResourcesStatusFacade.is_unique")
    @mock.patch("pcs.common.resource_status.ResourcesStatusFacade.get_type")
    def test_unique(
        self,
        mock_get_type: mock.Mock,
        mock_is_unique: mock.Mock,
        mock_print: mock.Mock,
    ):
        for unique in [False, True]:
            self.lib_command.reset_mock()
            mock_get_type.reset_mock()
            mock_is_unique.reset_mock()
            mock_print.reset_mock()

            self.lib_command.return_value = ResourcesStatusDto(
                [fixture_primitive_dto("resource", None)]
            )
            mock_get_type.return_value = ResourceType.CLONE
            mock_is_unique.return_value = unique
            with self.subTest(value=unique):
                with self.assertRaises(SystemExit) as cm:
                    self._call_cmd(["resource", "clone", "unique"])
                self.assertEqual(cm.exception.code, 0 if unique else 2)
                self.lib_command.assert_called_once_with()
                mock_get_type.assert_called_once_with("resource", None)
                mock_is_unique.assert_called_once_with("resource", None)
                mock_print.assert_called_once_with(unique)

    @mock.patch(
        "pcs.common.resource_status.ResourcesStatusFacade.is_promotable"
    )
    @mock.patch("pcs.common.resource_status.ResourcesStatusFacade.get_type")
    def test_promotable(
        self,
        mock_get_type: mock.Mock,
        mock_is_promotable: mock.Mock,
        mock_print: mock.Mock,
    ):
        for promotable in [False, True]:
            self.lib_command.reset_mock()
            mock_get_type.reset_mock()
            mock_is_promotable.reset_mock()
            mock_print.reset_mock()

            self.lib_command.return_value = ResourcesStatusDto(
                [fixture_primitive_dto("resource", None)]
            )
            mock_get_type.return_value = ResourceType.CLONE
            mock_is_promotable.return_value = promotable
            with self.subTest(value=promotable):
                with self.assertRaises(SystemExit) as cm:
                    self._call_cmd(["resource", "clone", "promotable"])
                self.assertEqual(cm.exception.code, 0 if promotable else 2)
                self.lib_command.assert_called_once_with()
                mock_get_type.assert_called_once_with("resource", None)
                mock_is_promotable.assert_called_once_with("resource", None)
                mock_print.assert_called_once_with(promotable)

    @mock.patch("pcs.common.resource_status.ResourcesStatusFacade.is_unique")
    @mock.patch(
        "pcs.common.resource_status.ResourcesStatusFacade.is_promotable"
    )
    @mock.patch("pcs.common.resource_status.ResourcesStatusFacade.get_type")
    def test_unique_promotable(
        self,
        mock_get_type: mock.Mock,
        mock_is_promotable: mock.Mock,
        mock_is_unique: mock.Mock,
        mock_print: mock.Mock,
    ):
        for unique in [False, True]:
            for promotable in [False, True]:
                self.lib_command.reset_mock()
                mock_get_type.reset_mock()
                mock_is_unique.reset_mock()
                mock_is_promotable.reset_mock()
                mock_print.reset_mock()

                self.lib_command.return_value = ResourcesStatusDto([])
                mock_get_type.return_value = ResourceType.CLONE
                mock_is_unique.return_value = unique
                mock_is_promotable.return_value = promotable
                with self.subTest(
                    value=f"unique:{unique} promotable:{promotable}"
                ):
                    with self.assertRaises(SystemExit) as cm:
                        self._call_cmd(
                            ["resource", "clone", "unique", "promotable"]
                        )

                    self.assertEqual(
                        cm.exception.code, 0 if unique and promotable else 2
                    )
                    self.lib_command.assert_called_once_with()
                    mock_get_type.assert_called_once_with("resource", None)
                    mock_is_unique.assert_called_once_with("resource", None)

                    if unique:
                        # will be called only if previous checks were true
                        mock_is_promotable.assert_called_once_with(
                            "resource", None
                        )
                    else:
                        mock_is_promotable.assert_not_called()

                    mock_print.assert_called_once_with(unique and promotable)

    def test_quiet(self, mock_print: mock.Mock):
        self.lib_command.return_value = ResourcesStatusDto(
            [fixture_primitive_dto("resource", None)]
        )
        with self.assertRaises(SystemExit) as cm:
            self._call_cmd(["resource", "primitive"], {"quiet": True})
        self.lib_command.assert_called_once_with()
        self.assertEqual(cm.exception.code, 0)
        mock_print.assert_not_called()


@mock.patch("pcs.cli.query.resource.print")
class TestQueryGetType(QueryBaseMixin, TestCase):
    def _call_cmd(self, argv, modifiers=None) -> None:
        modifiers = modifiers or {}
        resource.get_type(self.lib, argv, dict_to_modifiers(modifiers))

    def test_too_many_args(self, mock_print):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd(["foo", "bar"])
        self.assertIsNone(cm.exception.message)
        self.lib_command.assert_not_called()
        mock_print.assert_not_called()

    def test_nonexistent(self, mock_print: mock.Mock):
        self.nonexistent_should_fail(mock_print, [])

    def test_primitive(self, mock_print: mock.Mock):
        self.lib_command.return_value = ResourcesStatusDto(
            [fixture_primitive_dto("resource", None)]
        )
        self._call_cmd(["resource"])
        self.lib_command.assert_called_once_with()
        mock_print.assert_called_once_with("primitive")

    @mock.patch("pcs.common.resource_status.ResourcesStatusFacade.is_unique")
    @mock.patch(
        "pcs.common.resource_status.ResourcesStatusFacade.is_promotable"
    )
    @mock.patch("pcs.common.resource_status.ResourcesStatusFacade.get_type")
    def test_all_types(
        self,
        mock_get_type: mock.Mock,
        mock_is_promotable: mock.Mock,
        mock_is_unique: mock.Mock,
        mock_print: mock.Mock,
    ):
        for resource_type in ResourceType:
            self.lib_command.reset_mock()
            mock_get_type.reset_mock()
            mock_is_promotable.reset_mock()
            mock_is_unique.reset_mock()
            mock_print.reset_mock()

            self.lib_command.return_value = ResourcesStatusDto(
                [fixture_primitive_dto("resource", None)]
            )
            mock_get_type.return_value = resource_type
            mock_is_unique.return_value = False
            mock_is_promotable.return_value = False
            with self.subTest(value=resource_type.name.lower()):
                self._call_cmd(["resource"])

                self.lib_command.assert_called_once_with()
                mock_get_type.assert_called_once_with("resource", None)
                if can_be_unique(resource_type):
                    mock_is_unique.assert_called_once_with("resource", None)
                else:
                    mock_is_unique.assert_not_called()
                if can_be_promotable(resource_type):
                    mock_is_promotable.assert_called_once_with("resource", None)
                else:
                    mock_is_promotable.assert_not_called()
                mock_print.assert_called_once_with(resource_type.name.lower())

    @mock.patch("pcs.common.resource_status.ResourcesStatusFacade.is_unique")
    @mock.patch(
        "pcs.common.resource_status.ResourcesStatusFacade.is_promotable"
    )
    @mock.patch("pcs.common.resource_status.ResourcesStatusFacade.get_type")
    def test_promotable_unique(
        self,
        mock_get_type: mock.Mock,
        mock_is_promotable: mock.Mock,
        mock_is_unique: mock.Mock,
        mock_print: mock.Mock,
    ):
        for promotable in [False, True]:
            for unique in [False, True]:
                expected_return_value = "clone"
                if unique:
                    expected_return_value += " unique"
                if promotable:
                    expected_return_value += " promotable"

                self.lib_command.reset_mock()
                mock_get_type.reset_mock()
                mock_is_unique.reset_mock()
                mock_is_promotable.reset_mock()
                mock_print.reset_mock()

                self.lib_command.return_value = ResourcesStatusDto([])
                mock_is_unique.return_value = unique
                mock_is_promotable.return_value = promotable
                mock_get_type.return_value = ResourceType.CLONE

                with self.subTest(value=expected_return_value):
                    self._call_cmd(["resource"])

                    self.lib_command.assert_called_once_with()
                    mock_get_type.assert_called_once_with("resource", None)
                    mock_is_unique.assert_called_once_with("resource", None)
                    mock_is_promotable.assert_called_once_with("resource", None)
                    mock_print.assert_called_once_with(expected_return_value)


@mock.patch("pcs.cli.query.resource.print")
class TestQueryIsStonith(QueryBaseMixin, TestCase):
    def _call_cmd(self, argv, modifiers=None) -> None:
        modifiers = modifiers or {}
        resource.is_stonith(self.lib, argv, dict_to_modifiers(modifiers))

    def test_too_many_args(self, mock_print):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd(["foo", "bar"])
        self.assertIsNone(cm.exception.message)
        self.lib_command.assert_not_called()
        mock_print.assert_not_called()

    def test_nonexistent(self, mock_print: mock.Mock):
        self.nonexistent_should_fail(mock_print, [])

    def test_true(self, mock_print: mock.Mock):
        self.lib_command.return_value = ResourcesStatusDto(
            [
                fixture_primitive_dto(
                    "resource", None, resource_agent="stonith:fence_xvm"
                )
            ]
        )
        with self.assertRaises(SystemExit) as cm:
            self._call_cmd(["resource"])
        self.assertEqual(cm.exception.code, 0)
        self.lib_command.assert_called_once_with()
        mock_print.assert_called_once_with(True)

    def test_false(self, mock_print: mock.Mock):
        self.lib_command.return_value = ResourcesStatusDto(
            [fixture_primitive_dto("resource", None)]
        )
        with self.assertRaises(SystemExit) as cm:
            self._call_cmd(["resource"])
        self.assertEqual(cm.exception.code, 2)
        self.lib_command.assert_called_once_with()
        mock_print.assert_called_once_with(False)

    def test_quiet(self, mock_print: mock.Mock):
        self.lib_command.return_value = ResourcesStatusDto(
            [
                fixture_primitive_dto(
                    "resource", None, resource_agent="stonith:fence_xvm"
                )
            ]
        )
        with self.assertRaises(SystemExit) as cm:
            self._call_cmd(["resource"], {"quiet": True})
        self.assertEqual(cm.exception.code, 0)
        self.lib_command.assert_called_once_with()
        mock_print.assert_not_called()


@mock.patch("pcs.cli.query.resource.print")
class TestQueryGetMembers(QueryBaseMixin, TestCase):
    def _call_cmd(self, argv, modifiers=None) -> None:
        modifiers = modifiers or {}
        resource.get_members(self.lib, argv, dict_to_modifiers(modifiers))

    def test_too_many_args(self, mock_print):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd(["foo", "bar"])
        self.assertIsNone(cm.exception.message)
        self.lib_command.assert_not_called()
        mock_print.assert_not_called()

    def test_nonexistent(self, mock_print: mock.Mock):
        self.nonexistent_should_fail(mock_print, [])

    def test_bad_resource_type(self, mock_print: mock.Mock):
        self.lib_command.return_value = ResourcesStatusDto(
            [fixture_primitive_dto("resource", None)]
        )
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd(["resource"])
        self.assertEqual(
            cm.exception.message,
            (
                "Resource 'resource' has unexpected type 'primitive'. This "
                "command works only for resources of type 'bundle', 'clone', "
                "'group'"
            ),
        )
        self.lib_command.assert_called_once_with()
        mock_print.assert_not_called()

    def test_no_member(self, mock_print: mock.Mock):
        self.lib_command.return_value = ResourcesStatusDto(
            [fixture_group_dto("resource", None, [])]
        )
        self._call_cmd(["resource"])
        self.lib_command.assert_called_once_with()
        mock_print.assert_called_once_with("")

    def test_single_member(self, mock_print: mock.Mock):
        self.lib_command.return_value = ResourcesStatusDto(
            [
                fixture_group_dto(
                    "resource", None, [fixture_primitive_dto("a", None)]
                )
            ]
        )
        self._call_cmd(["resource"])
        self.lib_command.assert_called_once_with()
        mock_print.assert_called_once_with("a")

    def test_multiple_members(self, mock_print: mock.Mock):
        self.lib_command.return_value = ResourcesStatusDto(
            [
                fixture_group_dto(
                    "resource",
                    None,
                    [
                        fixture_primitive_dto("a", None),
                        fixture_primitive_dto("c", None),
                        fixture_primitive_dto("b", None),
                    ],
                )
            ]
        )
        self._call_cmd(["resource"])
        self.lib_command.assert_called_once_with()
        mock_print.assert_called_once_with("a\nc\nb")


@mock.patch("pcs.cli.query.resource.print")
class TestQueryGetNodes(QueryBaseMixin, TestCase):
    def _call_cmd(self, argv, modifiers=None) -> None:
        modifiers = modifiers or {}
        resource.get_nodes(self.lib, argv, dict_to_modifiers(modifiers))

    def test_nonexistent(self, mock_print: mock.Mock):
        self.nonexistent_should_fail(mock_print, [])

    def test_too_many_args(self, mock_print):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd(["foo", "bar"])
        self.assertIsNone(cm.exception.message)
        self.lib_command.assert_not_called()
        mock_print.assert_not_called()

    def test_no_node(self, mock_print: mock.Mock):
        self.lib_command.return_value = ResourcesStatusDto(
            [fixture_primitive_dto("resource", None, node_names=[])]
        )
        self._call_cmd(["resource"])
        self.lib_command.assert_called_once_with()
        mock_print.assert_called_once_with("")

    def test_one_node(self, mock_print: mock.Mock):
        self.lib_command.return_value = ResourcesStatusDto(
            [fixture_primitive_dto("resource", None)]
        )
        self._call_cmd(["resource"])
        self.lib_command.assert_called_once_with()
        mock_print.assert_called_once_with("node1")

    def test_multiple_nodes(self, mock_print: mock.Mock):
        self.lib_command.return_value = ResourcesStatusDto(
            [
                fixture_primitive_dto(
                    "resource", None, node_names=["node1", "node42", "node2"]
                )
            ]
        )
        self._call_cmd(["resource"])
        self.lib_command.assert_called_once_with()
        mock_print.assert_called_once_with("node1\nnode2\nnode42")


@mock.patch("pcs.cli.query.resource.print")
class TestQueryIsState(QueryBaseMixin, TestCase):
    def _call_cmd(self, argv, modifiers=None) -> None:
        modifiers = modifiers or {}
        resource.is_state(self.lib, argv, dict_to_modifiers(modifiers))

    def test_nonexistent(self, mock_print: mock.Mock):
        self.nonexistent_should_fail(mock_print, ["started"])

    def test_simple_true(self, mock_print: mock.Mock):
        self.lib_command.return_value = ResourcesStatusDto(
            [fixture_primitive_dto("resource", None)]
        )
        with self.assertRaises(SystemExit) as cm:
            self._call_cmd(["resource", "started"])
        self.assertEqual(cm.exception.code, 0)
        self.lib_command.assert_called_once_with()
        mock_print.assert_called_once_with(True)

    def test_simple_false(self, mock_print: mock.Mock):
        self.lib_command.return_value = ResourcesStatusDto(
            [fixture_primitive_dto("resource", None)]
        )
        with self.assertRaises(SystemExit) as cm:
            self._call_cmd(["resource", "stopped"])
        self.assertEqual(cm.exception.code, 2)
        self.lib_command.assert_called_once_with()
        mock_print.assert_called_once_with(False)

    def test_bad_state(self, mock_print: mock.Mock):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd(["resource", "start"])
        self.assertEqual(
            cm.exception.message,
            (
                "'start' is not a valid state value, use 'active', 'blocked', "
                "'demoting', 'disabled', 'enabled', 'failed', 'failure_ignored'"
                ", 'locked_to', 'maintenance', 'managed', 'migrating', "
                "'monitoring', 'orphaned', 'pending', 'promoted', 'promoting', "
                "'started', 'starting', 'stopped', 'stopping', 'unmanaged', "
                "'unpromoted'"
            ),
        )
        self.lib_command.assert_not_called()
        mock_print.assert_not_called()

    def test_quiet(self, mock_print: mock.Mock):
        self.lib_command.return_value = ResourcesStatusDto(
            [fixture_primitive_dto("resource", None)]
        )
        with self.assertRaises(SystemExit) as cm:
            self._call_cmd(["resource", "started"], {"quiet": True})
        self.assertEqual(cm.exception.code, 0)
        self.lib_command.assert_called_once_with()
        mock_print.assert_not_called()

    @mock.patch("pcs.common.resource_status.ResourcesStatusFacade.is_state")
    def test_all_states(self, mock_is_state: mock.Mock, mock_print: mock.Mock):
        state_list = [state.name.lower() for state in ResourceState]
        for state in state_list:
            self.lib_command.reset_mock()
            mock_is_state.reset_mock()
            mock_print.reset_mock()

            # provide empty status dto so we dont have to mock the whole facade
            self.lib_command.return_value = ResourcesStatusDto([])
            mock_is_state.return_value = True

            with self.subTest(value=state):
                with self.assertRaises(SystemExit) as cm:
                    self._call_cmd(["resource", state])

                self.assertEqual(cm.exception.code, 0)
                self.lib_command.assert_called_once_with()
                mock_is_state.assert_called_once_with(
                    "resource",
                    None,
                    ResourceState[state.upper()],
                    None,
                    None,
                    None,
                )
                mock_print.assert_called_once_with(True)

    def test_on_node_missing_node(self, mock_print: mock.Mock):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd(["resource", "started", "on-node"])
        self.assertIsNone(cm.exception.message)
        self.lib_command.assert_not_called()
        mock_print.assert_not_called()

    def test_on_node_too_much_nodes(self, mock_print: mock.Mock):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd(["resource", "started", "on-node", "node1", "node2"])
        self.assertIsNone(cm.exception.message)
        self.lib_command.assert_not_called()
        mock_print.assert_not_called()

    @mock.patch("pcs.common.resource_status.ResourcesStatusFacade.is_state")
    def test_on_node(self, mock_is_state: mock.Mock, mock_print: mock.Mock):
        # provide empty status dto so we dont have to mock the whole facade
        self.lib_command.return_value = ResourcesStatusDto([])
        mock_is_state.return_value = True
        with self.assertRaises(SystemExit) as cm:
            self._call_cmd(["resource", "started", "on-node", "node1"])
        self.assertEqual(cm.exception.code, 0)

        self.lib_command.assert_called_once_with()
        mock_is_state.assert_called_once_with(
            "resource", None, ResourceState.STARTED, "node1", None, None
        )
        mock_print.assert_called_once_with(True)

    def test_members_quantifier_missing(self, mock_print: mock.Mock):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd(["resource", "started", "members"])
        self.assertIsNone(cm.exception.message)
        self.lib_command.assert_not_called()
        mock_print.assert_not_called()

    def test_members_quantifier_bad_value(self, mock_print: mock.Mock):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd(["resource", "started", "members", "foo"])
        self.assertEqual(
            cm.exception.message,
            "'foo' is not a valid members value, use 'all', 'any', 'none'",
        )
        self.lib_command.assert_not_called()
        mock_print.assert_not_called()

    def test_instances_quantifier_bad_value(self, mock_print: mock.Mock):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd(["resource", "started", "instances", "foo"])
        self.assertEqual(
            cm.exception.message,
            "'foo' is not a valid instances value, use 'all', 'any', 'none'",
        )
        self.lib_command.assert_not_called()
        mock_print.assert_not_called()

    @mock.patch("pcs.common.resource_status.ResourcesStatusFacade.is_state")
    def test_all_quantifiers(
        self, mock_is_state: mock.Mock, mock_print: mock.Mock
    ):
        quantifier_list = [
            quantifier.name.lower() for quantifier in MoreChildrenQuantifierType
        ]

        for member_quantifier in quantifier_list:
            for instances_quantifier in quantifier_list:
                self.lib_command.reset_mock()
                mock_is_state.reset_mock()
                mock_print.reset_mock()

                # provide empty status dto so we dont have to mock the whole facade
                self.lib_command.return_value = ResourcesStatusDto([])
                mock_is_state.return_value = True

                with self.subTest(
                    value=f"members {member_quantifier} instances:{instances_quantifier}"
                ):
                    with self.assertRaises(SystemExit) as cm:
                        self._call_cmd(
                            [
                                "resource",
                                "started",
                                "members",
                                member_quantifier,
                                "instances",
                                instances_quantifier,
                            ]
                        )

                    self.assertEqual(cm.exception.code, 0)
                    self.lib_command.assert_called_once_with()
                    mock_is_state.assert_called_once_with(
                        "resource",
                        None,
                        ResourceState.STARTED,
                        None,
                        MoreChildrenQuantifierType[member_quantifier.upper()],
                        MoreChildrenQuantifierType[
                            instances_quantifier.upper()
                        ],
                    )
                    mock_print.assert_called_once_with(True)

    def test_bad_members_quantifier(self, mock_print: mock.Mock):
        self.lib_command.return_value = ResourcesStatusDto(
            [fixture_primitive_dto("resource", None)]
        )
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd(["resource", "started", "members", "all"])
        self.assertEqual(
            cm.exception.message,
            (
                "'members' quantifier can be used only on group resources or "
                "group instances of cloned groups"
            ),
        )
        self.lib_command.assert_called_once_with()
        mock_print.assert_not_called()

    def test_bad_instances_quantifier(self, mock_print: mock.Mock):
        self.lib_command.return_value = ResourcesStatusDto(
            [fixture_primitive_dto("resource", None)]
        )
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd(["resource", "started", "instances", "all"])
        self.assertEqual(
            cm.exception.message,
            (
                "'instances' quantifier can be used only on clone resources "
                "and their instances, or on bundle resources and their replicas"
            ),
        )
        self.lib_command.assert_called_once_with()
        mock_print.assert_not_called()

    @mock.patch("pcs.common.resource_status.ResourcesStatusFacade.is_state")
    def test_all_args(self, mock_is_state: mock.Mock, mock_print: mock.Mock):
        self.lib_command.return_value = ResourcesStatusDto([])
        mock_is_state.return_value = True

        with self.assertRaises(SystemExit) as cm:
            self._call_cmd(
                [
                    "resource:1",
                    "started",
                    "on-node",
                    "node1",
                    "members",
                    "all",
                    "instances",
                    "any",
                ]
            )

        self.assertEqual(cm.exception.code, 0)
        self.lib_command.assert_called_once_with()
        mock_is_state.assert_called_once_with(
            "resource",
            "1",
            ResourceState.STARTED,
            "node1",
            MoreChildrenQuantifierType.ALL,
            MoreChildrenQuantifierType.ANY,
        )
        mock_print.assert_called_once_with(True)

    @mock.patch("pcs.common.resource_status.ResourcesStatusFacade.is_state")
    def test_clone_and_bundle(
        self, mock_is_state: mock.Mock, mock_print: mock.Mock
    ):
        # provide empty status dto so we dont have to mock the whole facade
        self.lib_command.return_value = ResourcesStatusDto([])
        mock_is_state.side_effect = NotImplementedError("foo")
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd(["resource", "started"])

        self.assertEqual(cm.exception.message, "foo")
        self.lib_command.assert_called_once_with()
        mock_is_state.assert_called_once_with(
            "resource", None, ResourceState.STARTED, None, None, None
        )
        mock_print.assert_not_called()


class QueryInContainerBaseMixin(QueryBaseMixin):
    not_in_container_status = ResourcesStatusDto(
        [fixture_primitive_dto("resource", None)]
    )

    def test_nonexistent(self, mock_print: mock.Mock):
        self.nonexistent_should_fail(mock_print, [])

    def test_too_many_args(self, mock_print):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd(["resource", "container", "foobar"])

        self.assertIsNone(cm.exception.message)
        self.lib_command.assert_not_called()
        mock_print.assert_not_called()

    def test_true_no_container_id(self, mock_print: mock.Mock):
        self.lib_command.return_value = self.in_container_status

        with self.assertRaises(SystemExit) as cm:
            self._call_cmd(["resource"])

        self.assertEqual(cm.exception.code, 0)
        self.lib_command.assert_called_once_with()
        calls = (mock.call(True), mock.call("container"))
        self.assertEqual(mock_print.call_count, len(calls))
        mock_print.assert_has_calls(calls)

    def test_false_no_container_id(self, mock_print: mock.Mock):
        self.lib_command.return_value = self.not_in_container_status

        with self.assertRaises(SystemExit) as cm:
            self._call_cmd(["resource"])

        self.assertEqual(cm.exception.code, 2)
        self.lib_command.assert_called_once_with()
        mock_print.assert_called_once_with(False)

    def test_true_container_id(self, mock_print: mock.Mock):
        self.lib_command.return_value = self.in_container_status

        with self.assertRaises(SystemExit) as cm:
            self._call_cmd(["resource", "container"])

        self.assertEqual(cm.exception.code, 0)
        self.lib_command.assert_called_once_with()
        calls = (mock.call(True), mock.call("container"))
        self.assertEqual(mock_print.call_count, len(calls))
        mock_print.assert_has_calls(calls)

    def test_true_bad_container_id(self, mock_print: mock.Mock):
        self.lib_command.return_value = self.in_container_status

        with self.assertRaises(SystemExit) as cm:
            self._call_cmd(["resource", "not_the_same_container"])

        self.assertEqual(cm.exception.code, 2)
        self.lib_command.assert_called_once_with()
        calls = (mock.call(False), mock.call("container"))
        self.assertEqual(mock_print.call_count, len(calls))
        mock_print.assert_has_calls(calls)

    def test_true_quiet(self, mock_print: mock.Mock):
        self.lib_command.return_value = self.in_container_status

        with self.assertRaises(SystemExit) as cm:
            self._call_cmd(["resource"], {"quiet": True})

        self.assertEqual(cm.exception.code, 0)
        self.lib_command.assert_called_once_with()
        mock_print.assert_not_called()

    def test_false_quiet(self, mock_print: mock.Mock):
        self.lib_command.return_value = self.not_in_container_status

        with self.assertRaises(SystemExit) as cm:
            self._call_cmd(["resource"], {"quiet": True})

        self.assertEqual(cm.exception.code, 2)
        self.lib_command.assert_called_once_with()
        mock_print.assert_not_called()


@mock.patch("pcs.cli.query.resource.print")
class TestQueryIsInGroup(QueryInContainerBaseMixin, TestCase):
    in_container_status = ResourcesStatusDto(
        [
            fixture_group_dto(
                "container", None, [fixture_primitive_dto("resource", None)]
            )
        ]
    )

    def _call_cmd(self, argv, modifiers=None) -> None:
        modifiers = modifiers or {}
        resource.is_in_group(self.lib, argv, dict_to_modifiers(modifiers))

    def test_bad_type(self, mock_print: mock.Mock):
        self.lib_command.return_value = self.in_container_status

        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd(["container"])

        self.assertEqual(
            cm.exception.message,
            (
                "Resource 'container' has unexpected type 'group'. This command"
                " works only for resources of type 'primitive'"
            ),
        )
        self.lib_command.assert_called_once_with()
        mock_print.assert_not_called()


@mock.patch("pcs.cli.query.resource.print")
class TestQueryIsInClone(QueryInContainerBaseMixin, TestCase):
    in_container_status = ResourcesStatusDto(
        [
            CloneStatusDto(
                "container",
                False,
                False,
                False,
                None,
                True,
                False,
                False,
                False,
                None,
                [fixture_primitive_dto("resource", None)],
            )
        ]
    )

    def _call_cmd(self, argv, modifiers=None) -> None:
        modifiers = modifiers or {}
        resource.is_in_clone(self.lib, argv, dict_to_modifiers(modifiers))

    def test_bad_type(self, mock_print: mock.Mock):
        self.lib_command.return_value = self.in_container_status

        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd(["container"])

        self.assertEqual(
            cm.exception.message,
            (
                "Resource 'container' has unexpected type 'clone'. This "
                "command works only for resources of type 'group', 'primitive'"
            ),
        )
        self.lib_command.assert_called_once_with()
        mock_print.assert_not_called()


@mock.patch("pcs.cli.query.resource.print")
class TestQueryIsInBundle(QueryInContainerBaseMixin, TestCase):
    in_container_status = ResourcesStatusDto(
        [
            BundleStatusDto(
                "container",
                "podman",
                "",
                False,
                False,
                None,
                True,
                False,
                [
                    BundleReplicaStatusDto(
                        "0",
                        member=fixture_primitive_dto("resource", None),
                        remote=None,
                        container=fixture_primitive_dto("podman", None),
                        ip_address=None,
                    )
                ],
            )
        ]
    )

    def _call_cmd(self, argv, modifiers=None) -> None:
        modifiers = modifiers or {}
        resource.is_in_bundle(self.lib, argv, dict_to_modifiers(modifiers))

    def test_bad_type(self, mock_print: mock.Mock):
        self.lib_command.return_value = self.in_container_status

        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd(["container"])

        self.assertEqual(
            cm.exception.message,
            (
                "Resource 'container' has unexpected type 'bundle'. This "
                "command works only for resources of type 'primitive'"
            ),
        )
        self.lib_command.assert_called_once_with()
        mock_print.assert_not_called()


@mock.patch("pcs.cli.query.resource.print")
class TestQueryGetIndexInGroup(QueryBaseMixin, TestCase):
    def _call_cmd(self, argv, modifiers=None) -> None:
        modifiers = modifiers or {}
        resource.get_index_in_group(
            self.lib, argv, dict_to_modifiers(modifiers)
        )

    def test_nonexistent(self, mock_print: mock.Mock):
        self.nonexistent_should_fail(mock_print, [])

    def test_too_many_args(self, mock_print):
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd(["resource", "foobar"])
        self.assertIsNone(cm.exception.message)
        self.lib_command.assert_not_called()
        mock_print.assert_not_called()

    def test_in_group(self, mock_print: mock.Mock):
        self.lib_command.return_value = ResourcesStatusDto(
            [
                fixture_group_dto(
                    "container", None, [fixture_primitive_dto("resource", None)]
                )
            ]
        )
        self._call_cmd(["resource"])

        self.lib_command.assert_called_once_with()
        mock_print.assert_called_once_with(0)

    def test_not_in_group(self, mock_print: mock.Mock):
        self.lib_command.return_value = ResourcesStatusDto(
            [fixture_primitive_dto("resource", None)]
        )
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd(["resource"])
        self.assertEqual(
            cm.exception.message, "Resource 'resource' is not in a group"
        )
        self.lib_command.assert_called_once_with()
        mock_print.assert_not_called()

    def test_bad_type(self, mock_print: mock.Mock):
        self.lib_command.return_value = ResourcesStatusDto(
            [fixture_group_dto("resource", None, [])]
        )
        with self.assertRaises(CmdLineInputError) as cm:
            self._call_cmd(["resource"])
        self.assertEqual(
            cm.exception.message,
            (
                "Resource 'resource' has unexpected type 'group'. This command "
                "works only for resources of type 'primitive'"
            ),
        )
        self.lib_command.assert_called_once_with()
        mock_print.assert_not_called()



import json
from shlex import split
from unittest import TestCase

from pcs.common.interface.dto import to_dict
from pcs.common.types import CibRuleInEffectStatus
from pcs.lib.cib.tools import get_resources

from pcs_test.tools import fixture_cib
from pcs_test.tools.constraints_dto import get_all_constraints
from pcs_test.tools.custom_mock import RuleInEffectEvalMock
from pcs_test.tools.misc import (
    get_test_resource,
    get_tmp_file,
    outdent,
    write_data_to_tmpfile,
)
from pcs_test.tools.pcs_runner import PcsRunner
from pcs_test.tools.xml import (
    XmlManipulation,
    etree_to_str,
)

RULE_EVAL = RuleInEffectEvalMock(
    {
        "loc_constr_with_expired_rule-rule": CibRuleInEffectStatus.EXPIRED,
        "loc_constr_with_not_expired_rule-rule": CibRuleInEffectStatus.IN_EFFECT,
        "loc_constr_with_not_expired_rule-1-rule": CibRuleInEffectStatus.IN_EFFECT,
    }
)


class ConstraintConfigJson(TestCase):
    def setUp(self):
        self.maxDiff = None
        self.pcs_runner = PcsRunner(
            cib_file=get_test_resource("cib-all.xml"),
        )

    def test_all_option(self):
        stdout, stderr, retval = self.pcs_runner.run(
            ["constraint", "config", "--output-format=json", "--all"]
        )
        expected = get_all_constraints(RULE_EVAL, include_expired=True)
        self.assertEqual(
            json.loads(stdout), json.loads(json.dumps(to_dict(expected)))
        )
        self.assertEqual(stderr, "")
        self.assertEqual(retval, 0)

    def test_success(self):
        stdout, stderr, retval = self.pcs_runner.run(
            ["constraint", "config", "--output-format=json"]
        )
        expected = get_all_constraints(RULE_EVAL, include_expired=False)
        self.assertEqual(
            json.loads(stdout), json.loads(json.dumps(to_dict(expected)))
        )
        self.assertEqual(stderr, "")
        self.assertEqual(retval, 0)

    def test_full_option(self):
        stdout, stderr, retval = self.pcs_runner.run(
            ["constraint", "config", "--output-format=json", "--full"]
        )
        self.assertEqual(stdout, "")
        self.assertEqual(
            stderr,
            "Error: Option '--full' is not compatible with 'json' output format.\n",
        )
        self.assertEqual(retval, 1)


class ConstraintConfigCmdMixin:
    orig_cib_file_path = get_test_resource("cib-all.xml")

    def setUp(self):
        self.new_cib_file = get_tmp_file(self._get_tmp_file_name())
        self.pcs_runner_orig = PcsRunner(cib_file=self.orig_cib_file_path)
        self.pcs_runner_new = PcsRunner(cib_file=self.new_cib_file.name)
        write_data_to_tmpfile(
            fixture_cib.modify_cib_file(
                get_test_resource("cib-empty.xml"),
                resources=etree_to_str(
                    get_resources(
                        XmlManipulation.from_file(self.orig_cib_file_path).tree
                    )
                ),
            ),
            self.new_cib_file,
        )
        self.maxDiff = None

    def tearDown(self):
        self.new_cib_file.close()

    def _get_as_json(self, runner, use_all):
        cmd = ["constraint", "config", "--output-format=json"]
        if use_all:
            cmd.append("--all")
        stdout, stderr, retval = runner.run(cmd)
        self.assertEqual(stderr, "")
        self.assertEqual(retval, 0)
        return json.loads(stdout)

    def _test_success(self, use_all):
        cmd = ["constraint", "config", "--output-format=cmd"]
        if use_all:
            cmd.append("--all")
        stdout, stderr, retval = self.pcs_runner_orig.run(cmd)
        self.assertEqual(retval, 0)
        cmds = [
            split(cmd)[1:]
            for cmd in stdout.replace("\\\n", "").strip().split(";\n")
        ]
        for cmd in cmds:
            stdout, stderr, retval = self.pcs_runner_new.run(cmd)
            self.assertEqual(
                retval,
                0,
                (
                    f"Command {cmd} exited with {retval}\nstdout:\n{stdout}\n"
                    f"stderr:\n{stderr}"
                ),
            )
        self.assertEqual(
            self._get_as_json(self.pcs_runner_new, use_all),
            self._get_as_json(self.pcs_runner_orig, use_all),
        )

    def test_all(self):
        self._test_success(True)

    def test_not_all(self):
        self._test_success(False)


class ConstraintConfigCmd(ConstraintConfigCmdMixin, TestCase):
    @staticmethod
    def _get_tmp_file_name():
        return "tier1_constraint_test_config_cib.xml"


class ConstraintConfigCmdSpaceInDate(ConstraintConfigCmdMixin, TestCase):
    # This class tests that pcs exports dates from location rules constraint
    # with spaces replaced by T in pcs commands, so that they can be run and
    # processed by pcs correctly.
    orig_cib_file_path = get_test_resource("cib-rule-with-spaces-in-date.xml")

    @staticmethod
    def _get_tmp_file_name():
        return "tier1_constraint_test_config_cib_date_space.xml"

    @staticmethod
    def _replace(struct, search_replace):
        if isinstance(struct, dict):
            for key, val in struct.items():
                struct[key] = ConstraintConfigCmdSpaceInDate._replace(
                    val, search_replace
                )
            return struct
        if isinstance(struct, list):
            return [
                ConstraintConfigCmdSpaceInDate._replace(val, search_replace)
                for val in struct
            ]
        for search, replace in search_replace:
            if struct == search:
                return replace
        return struct

    def _get_as_json(self, runner, use_all):
        data = super()._get_as_json(runner, use_all)
        return self._replace(
            data,
            [
                ("2023-01-01 12:00", "2023-01-01T12:00"),
                ("2023-12-31 12:00", "2023-12-31T12:00"),
            ],
        )

    def test_commands(self):
        stdout, stderr, retval = self.pcs_runner_orig.run(
            ["constraint", "config", "--output-format=cmd"]
        )
        self.assertEqual(retval, 0)
        self.assertEqual(stderr, "")
        self.assertEqual(
            stdout,
            (
                "pcs -- constraint location resource%R1 rule \\\n"
                "  id=location-R1-rule constraint-id=location-R1 score=INFINITY \\\n"
                "  '#uname eq node1 and date gt 2023-01-01T12:00 and "
                "date lt 2023-12-31T12:00 and date in_range 2023-01-01T12:00 "
                "to 2023-12-31T12:00'\n"
            ),
        )


class ConstraintConfigCmdMultipleRulesInConstraint(
    ConstraintConfigCmdMixin, TestCase
):
    # This class tests that pcs exports a constraint containing more than one
    # rule. This configuration is no longer supported by pacemaker and thus pcs
    # cannot create it. So instead of 'pcs constraint rule add', the extra
    # rules are exported as separate constraints. Json export, however, matches
    # the CIB structure.
    orig_cib_file_path = get_test_resource(
        "cib-rule-several-times-in-constraint.xml"
    )

    @staticmethod
    def _get_tmp_file_name():
        return "tier1_constraint_test_config_cib_multiple_rules.xml"

    def test_commands(self):
        stdout, stderr, retval = self.pcs_runner_orig.run(
            ["constraint", "config", "--output-format=cmd"]
        )
        self.assertEqual(retval, 0)
        self.assertEqual(
            stderr,
            (
                "Warning: Constraint 'location-R1' contains more than one "
                "rule, which is no longer supported. Instead, each rule will "
                "be put in a separate constraint.\n"
            ),
        )
        self.assertEqual(
            stdout,
            (
                "pcs -- constraint location resource%R1 rule \\\n"
                "  id=location-R1-rule constraint-id=location-R1 score=INFINITY \\\n"
                "  '#uname eq node1 and date gt 2023-01-01T12:00';\n"
                "pcs -- constraint location resource%R1 rule \\\n"
                "  id=location-R1-rule-1 score=INFINITY \\\n"
                "  '#uname eq node2 and date gt 2023-01-01T12:00';\n"
                "pcs -- constraint location resource%R1 rule \\\n"
                "  id=location-R1-rule-2 score=INFINITY \\\n"
                "  '#uname eq node3 and date gt 2023-01-01T12:00'\n"
            ),
        )

    def test_all(self):
        # Multiple rules in a single constraint are transformed to multimple
        # constraints with a single rule. Therefore the json exports don't
        # match and so this test is disabled. We only test export to commands.
        pass

    def test_not_all(self):
        # Multiple rules in a single constraint are transformed to multimple
        # constraints with a single rule. Therefore the json exports don't
        # match and so this test is disabled. We only test export to commands.
        pass


class ConstraintConfigCmdUnsupported(TestCase):
    def setUp(self):
        self.maxDiff = None
        self.pcs_runner = PcsRunner(
            cib_file=get_test_resource("cib-unexportable-constraints.xml"),
        )

    def test_dont_export_unsupported_constraints(self):
        stdout, stderr, retval = self.pcs_runner.run(
            ["constraint", "config", "--output-format=cmd"]
        )
        self.assertEqual(retval, 0)
        suffix = "not supported by this command. Command for creating the constraint is omitted.\n"
        self.assertEqual(
            stderr,
            (
                f"Warning: Location set constraint with id 'location-set' configured but it's {suffix}"
                f"Warning: Resource role detected in constraint 'location-role' but {suffix}"
                f"Warning: Lifetime configuration detected in constraint 'location-lifetime' but {suffix}"
                f"Warning: Option 'influence' detected in constraint 'colocation-influence' but {suffix}"
                f"Warning: Lifetime configuration detected in constraint 'colocation-lifetime' but {suffix}"
                f"Warning: Option 'node-attribute' detected in constraint 'colocation-node-attribute' but {suffix}"
                f"Warning: Option 'ordering' detected in resource set 'colocation-set-ordering-set' but {suffix}"
                f"Warning: Option 'require-all' detected in constraint 'order-set-require-all' but {suffix}"
                f"Warning: Option 'ordering' detected in resource set 'order-set-ordering-set' but {suffix}"
            ),
        )
        self.assertEqual(
            stdout,
            (
                "pcs -- constraint location add location-OK resource%R1 node1 score=INFINITY;\n"
                "pcs -- constraint colocation add R1 with R3 \\\n"
                "  id=colocation-OK score=INFINITY;\n"
                "pcs -- constraint colocation \\\n"
                "  set R1 R3 \\\n"
                "  setoptions id=colocation-set-OK;\n"
                "pcs -- constraint order start R1 then start R3 \\\n"
                "  id=order-OK;\n"
                "pcs -- constraint order start R1 then start R3 \\\n"
                "  id=order-lifetime;\n"
                "pcs -- constraint order \\\n"
                "  set R1 R3 \\\n"
                "  setoptions id=order-set-OK\n"
            ),
        )


class ConstraintConfigText(TestCase):
    def setUp(self):
        self.maxDiff = None
        self.pcs_runner = PcsRunner(
            cib_file=get_test_resource("cib-all.xml"),
        )

    def test_success(self):
        stdout, stderr, retval = self.pcs_runner.run(["constraint", "config"])
        expected = outdent(
            """\
            Location Constraints:
              resource 'R7' avoids node 'non-existing-node' with score 10000
              resource 'R7' avoids node 'another-one' with score INFINITY
              resource 'R7' prefers node 'localhost' with score INFINITY
                resource-discovery=always
              resource 'G2' prefers node 'localhost' with score INFINITY
              resource pattern 'R*' prefers node 'localhost' with score INFINITY
              resource 'R6-clone'
                Rules:
                  Rule: boolean-op=and role=Unpromoted score=500
                    Expression: #uname eq node1
                    Expression: date gt 2000-01-01
              resource 'R6-clone'
                Rules:
                  Rule: boolean-op=and role=Promoted score-attribute=test-attr
                    Expression: date gt 2010-12-31
                    Expression: #uname eq node1
            Colocation Constraints:
              Promoted resource 'G1-clone' with Stopped resource 'R6-clone'
                score=-100
            Colocation Set Constraints:
              Set Constraint:
                score=-1
                Resource Set:
                  Resources: 'G2', 'R7'
                  role=Started
                Resource Set:
                  Resources: 'B2', 'R6-clone'
                  sequential=0
            Order Constraints:
              stop resource 'R7' then stop resource 'G2'
                symmetrical=0 require-all=0 score=-123
              start resource 'G2' then start resource 'B2'
                kind=Optional
            Order Set Constraints:
              Set Constraint:
                kind=Optional
                Resource Set:
                  Resources: 'B2', 'R6-clone'
                  require-all=0 action=stop
                Resource Set:
                  Resources: 'G1-clone'
                  sequential=0 action=promote
            Ticket Constraints:
              Promoted resource 'G1-clone' depends on ticket 'custom-ticket1'
                loss-policy=demote
            Ticket Set Constraints:
              Set Constraint:
                ticket=ticket2
                Resource Set:
                  Resources: 'B2', 'G2', 'R7'
                  role=Stopped
            """
        )
        self.assertEqual(stdout, expected)
        self.assertEqual(stderr, "")
        self.assertEqual(retval, 0)

    def test_all_option(self):
        stdout, stderr, retval = self.pcs_runner.run(
            ["constraint", "config", "--all"]
        )
        expected = outdent(
            """\
            Location Constraints:
              resource 'R7' avoids node 'non-existing-node' with score 10000
              resource 'R7' avoids node 'another-one' with score INFINITY
              resource 'R7' prefers node 'localhost' with score INFINITY
                resource-discovery=always
              resource 'G2' prefers node 'localhost' with score INFINITY
              resource pattern 'R*' prefers node 'localhost' with score INFINITY
              resource 'B2'
                Rules:
                  Rule (expired): boolean-op=and score=500
                    Expression: date lt 2000-01-01
              resource 'R6-clone'
                Rules:
                  Rule: boolean-op=and role=Unpromoted score=500
                    Expression: #uname eq node1
                    Expression: date gt 2000-01-01
              resource 'R6-clone'
                Rules:
                  Rule: boolean-op=and role=Promoted score-attribute=test-attr
                    Expression: date gt 2010-12-31
                    Expression: #uname eq node1
            Colocation Constraints:
              Promoted resource 'G1-clone' with Stopped resource 'R6-clone'
                score=-100
            Colocation Set Constraints:
              Set Constraint:
                score=-1
                Resource Set:
                  Resources: 'G2', 'R7'
                  role=Started
                Resource Set:
                  Resources: 'B2', 'R6-clone'
                  sequential=0
            Order Constraints:
              stop resource 'R7' then stop resource 'G2'
                symmetrical=0 require-all=0 score=-123
              start resource 'G2' then start resource 'B2'
                kind=Optional
            Order Set Constraints:
              Set Constraint:
                kind=Optional
                Resource Set:
                  Resources: 'B2', 'R6-clone'
                  require-all=0 action=stop
                Resource Set:
                  Resources: 'G1-clone'
                  sequential=0 action=promote
            Ticket Constraints:
              Promoted resource 'G1-clone' depends on ticket 'custom-ticket1'
                loss-policy=demote
            Ticket Set Constraints:
              Set Constraint:
                ticket=ticket2
                Resource Set:
                  Resources: 'B2', 'G2', 'R7'
                  role=Stopped
            """
        )
        self.assertEqual(stdout, expected)
        self.assertEqual(stderr, "")
        self.assertEqual(retval, 0)

    def test_full_option(self):
        stdout, stderr, retval = self.pcs_runner.run(
            ["constraint", "config", "--full"]
        )
        expected = outdent(
            """\
            Location Constraints:
              resource 'R7' avoids node 'non-existing-node' with score 10000 (id: location-R7-non-existing-node--10000)
              resource 'R7' avoids node 'another-one' with score INFINITY (id: location-R7-another-one--INFINITY)
              resource 'R7' prefers node 'localhost' with score INFINITY (id: location-R7-localhost-INFINITY)
                resource-discovery=always
              resource 'G2' prefers node 'localhost' with score INFINITY (id: location-G2-localhost-INFINITY)
              resource pattern 'R*' prefers node 'localhost' with score INFINITY (id: location-R-localhost-INFINITY)
              resource 'R6-clone' (id: loc_constr_with_not_expired_rule)
                Rules:
                  Rule: boolean-op=and role=Unpromoted score=500 (id: loc_constr_with_not_expired_rule-rule)
                    Expression: #uname eq node1 (id: loc_constr_with_not_expired_rule-rule-expr)
                    Expression: date gt 2000-01-01 (id: loc_constr_with_not_expired_rule-rule-expr-1)
              resource 'R6-clone' (id: loc_constr_with_not_expired_rule-1)
                Rules:
                  Rule: boolean-op=and role=Promoted score-attribute=test-attr (id: loc_constr_with_not_expired_rule-1-rule)
                    Expression: date gt 2010-12-31 (id: loc_constr_with_not_expired_rule-1-rule-expr)
                    Expression: #uname eq node1 (id: loc_constr_with_not_expired_rule-1-rule-expr-1)
            Colocation Constraints:
              Promoted resource 'G1-clone' with Stopped resource 'R6-clone' (id: colocation-G1-clone-R6-clone--100)
                score=-100
            Colocation Set Constraints:
              Set Constraint: colocation_set_R7G2B2
                score=-1
                Resource Set: colocation_set_R7G2B2_set
                  Resources: 'G2', 'R7'
                  role=Started
                Resource Set: colocation_set_R7G2B2_set-1
                  Resources: 'B2', 'R6-clone'
                  sequential=0
            Order Constraints:
              stop resource 'R7' then stop resource 'G2' (id: order-R7-G2-mandatory)
                symmetrical=0 require-all=0 score=-123
              start resource 'G2' then start resource 'B2' (id: order-G2-B2-Optional)
                kind=Optional
            Order Set Constraints:
              Set Constraint: order_set_B2R6-cloneSe
                kind=Optional
                Resource Set: order_set_B2R6-cloneSe_set
                  Resources: 'B2', 'R6-clone'
                  require-all=0 action=stop
                Resource Set: order_set_B2R6-cloneSe_set-1
                  Resources: 'G1-clone'
                  sequential=0 action=promote
            Ticket Constraints:
              Promoted resource 'G1-clone' depends on ticket 'custom-ticket1' (id: ticket-custom-ticket1-G1-clone-Promoted)
                loss-policy=demote
            Ticket Set Constraints:
              Set Constraint: ticket_set_R7B2G2
                ticket=ticket2
                Resource Set: ticket_set_R7B2G2_set
                  Resources: 'B2', 'G2', 'R7'
                  role=Stopped
            """
        )
        self.assertEqual(stdout, expected)
        self.assertEqual(stderr, "")
        self.assertEqual(retval, 0)

    def test_all_full_options(self):
        stdout, stderr, retval = self.pcs_runner.run(
            ["constraint", "config", "--full", "--all"]
        )
        expected = outdent(
            """\
            Location Constraints:
              resource 'R7' avoids node 'non-existing-node' with score 10000 (id: location-R7-non-existing-node--10000)
              resource 'R7' avoids node 'another-one' with score INFINITY (id: location-R7-another-one--INFINITY)
              resource 'R7' prefers node 'localhost' with score INFINITY (id: location-R7-localhost-INFINITY)
                resource-discovery=always
              resource 'G2' prefers node 'localhost' with score INFINITY (id: location-G2-localhost-INFINITY)
              resource pattern 'R*' prefers node 'localhost' with score INFINITY (id: location-R-localhost-INFINITY)
              resource 'B2' (id: loc_constr_with_expired_rule)
                Rules:
                  Rule (expired): boolean-op=and score=500 (id: loc_constr_with_expired_rule-rule)
                    Expression: date lt 2000-01-01 (id: loc_constr_with_expired_rule-rule-expr)
              resource 'R6-clone' (id: loc_constr_with_not_expired_rule)
                Rules:
                  Rule: boolean-op=and role=Unpromoted score=500 (id: loc_constr_with_not_expired_rule-rule)
                    Expression: #uname eq node1 (id: loc_constr_with_not_expired_rule-rule-expr)
                    Expression: date gt 2000-01-01 (id: loc_constr_with_not_expired_rule-rule-expr-1)
              resource 'R6-clone' (id: loc_constr_with_not_expired_rule-1)
                Rules:
                  Rule: boolean-op=and role=Promoted score-attribute=test-attr (id: loc_constr_with_not_expired_rule-1-rule)
                    Expression: date gt 2010-12-31 (id: loc_constr_with_not_expired_rule-1-rule-expr)
                    Expression: #uname eq node1 (id: loc_constr_with_not_expired_rule-1-rule-expr-1)
            Colocation Constraints:
              Promoted resource 'G1-clone' with Stopped resource 'R6-clone' (id: colocation-G1-clone-R6-clone--100)
                score=-100
            Colocation Set Constraints:
              Set Constraint: colocation_set_R7G2B2
                score=-1
                Resource Set: colocation_set_R7G2B2_set
                  Resources: 'G2', 'R7'
                  role=Started
                Resource Set: colocation_set_R7G2B2_set-1
                  Resources: 'B2', 'R6-clone'
                  sequential=0
            Order Constraints:
              stop resource 'R7' then stop resource 'G2' (id: order-R7-G2-mandatory)
                symmetrical=0 require-all=0 score=-123
              start resource 'G2' then start resource 'B2' (id: order-G2-B2-Optional)
                kind=Optional
            Order Set Constraints:
              Set Constraint: order_set_B2R6-cloneSe
                kind=Optional
                Resource Set: order_set_B2R6-cloneSe_set
                  Resources: 'B2', 'R6-clone'
                  require-all=0 action=stop
                Resource Set: order_set_B2R6-cloneSe_set-1
                  Resources: 'G1-clone'
                  sequential=0 action=promote
            Ticket Constraints:
              Promoted resource 'G1-clone' depends on ticket 'custom-ticket1' (id: ticket-custom-ticket1-G1-clone-Promoted)
                loss-policy=demote
            Ticket Set Constraints:
              Set Constraint: ticket_set_R7B2G2
                ticket=ticket2
                Resource Set: ticket_set_R7B2G2_set
                  Resources: 'B2', 'G2', 'R7'
                  role=Stopped
            """
        )
        self.assertEqual(stdout, expected)
        self.assertEqual(stderr, "")
        self.assertEqual(retval, 0)



from typing import Collection, NewType

AddRemoveContainerType = NewType("AddRemoveContainerType", str)
AddRemoveItemType = NewType("AddRemoveItemType", str)
BoothConfigUsedWhere = NewType("BoothConfigUsedWhere", str)
DefaultAddressSource = NewType("DefaultAddressSource", str)
FenceHistoryCommandType = NewType("FenceHistoryCommandType", str)
ForceCode = NewType("ForceCode", str)
# Container would suffice. However, this type is used in interfaces of many
# library commands. When calling the commands via API, their input comes in
# JSON and is transformed by dacite. In JSON, force flags are held in a list,
# there is no other way. And dacite is unable to deserialize a list to
# attributes annotated as Iterable or Container.
ForceFlags = Collection[ForceCode]
MessageCode = NewType("MessageCode", str)
DeprecatedMessageCode = NewType("DeprecatedMessageCode", str)
PcsCommand = NewType("PcsCommand", str)
ReasonType = NewType("ReasonType", str)
ServiceAction = NewType("ServiceAction", str)
SeverityLevel = NewType("SeverityLevel", str)
StonithRestartlessUpdateUnableToPerformReason = NewType(
    "StonithRestartlessUpdateUnableToPerformReason", str
)
StonithWatchdogTimeoutCannotBeSetReason = NewType(
    "StonithWatchdogTimeoutCannotBeSetReason", str
)



from unittest import TestCase

from pcs.common.reports import codes as report_codes
from pcs.lib.corosync import config_validators

from pcs_test.tools import fixture
from pcs_test.tools.assertions import assert_report_item_list_equal


class RenameCluster(TestCase):
    # pylint: disable=no-self-use
    def test_valid_cluster_name(self):
        assert_report_item_list_equal(
            config_validators.rename_cluster("my-cluster", False), []
        )

    def test_empty_name(self):
        assert_report_item_list_equal(
            config_validators.rename_cluster(""),
            [
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_name="cluster name",
                    option_value="",
                    allowed_values=None,
                    cannot_be_empty=True,
                    forbidden_characters=None,
                )
            ],
        )

    def test_gfs2_too_long(self):
        assert_report_item_list_equal(
            config_validators.rename_cluster(33 * "a"),
            [
                fixture.error(
                    report_codes.COROSYNC_CLUSTER_NAME_INVALID_FOR_GFS2,
                    force_code=report_codes.FORCE,
                    cluster_name=(33 * "a"),
                    max_length=32,
                    allowed_characters="a-z A-Z 0-9 _-",
                ),
            ],
        )

    def test_gfs2_bad_characters(self):
        assert_report_item_list_equal(
            config_validators.rename_cluster("cluster.name"),
            [
                fixture.error(
                    report_codes.COROSYNC_CLUSTER_NAME_INVALID_FOR_GFS2,
                    force_code=report_codes.FORCE,
                    cluster_name="cluster.name",
                    max_length=32,
                    allowed_characters="a-z A-Z 0-9 _-",
                ),
            ],
        )

    def test_gfs2_forced(self):
        cluster_name = (16 * "a") + ".: @" + (16 * "b")
        assert_report_item_list_equal(
            config_validators.rename_cluster(
                cluster_name, force_cluster_name=True
            ),
            [
                fixture.warn(
                    report_codes.COROSYNC_CLUSTER_NAME_INVALID_FOR_GFS2,
                    cluster_name=cluster_name,
                    max_length=32,
                    allowed_characters="a-z A-Z 0-9 _-",
                ),
            ],
        )



from textwrap import dedent
from unittest import TestCase

from lxml import etree

from pcs.common.str_tools import format_list, format_plural

from pcs_test.tools.assertions import AssertPcsMixin
from pcs_test.tools.cib import get_assert_pcs_effect_mixin
from pcs_test.tools.fixture_cib import modify_cib_file
from pcs_test.tools.misc import (
    get_test_resource,
    get_tmp_file,
    write_data_to_tmpfile,
)
from pcs_test.tools.pcs_runner import PcsRunner


def fixture_primitive_xml(resource_id: str) -> str:
    return f"""
        <primitive id="{resource_id}" class="ocf" provider="pcsmock" type="minimal"/>
    """


FIXTURE_CLONED_GROUP_XML = f"""
    <clone id="R2-clone">
        <group id="R2-group">
            {fixture_primitive_xml("R2")}
        </group>
    </clone>
"""
FIXTURE_GROUP_XML = f"""
    <group id="R3R4-group">
        {fixture_primitive_xml("R3")}
        {fixture_primitive_xml("R4")}
    </group>
"""
FIXTURE_TAG_XML = """
    <tag id="TAG">
        <obj_ref id="R3"/>
        <obj_ref id="R4"/>
    </tag>
"""
FIXTURE_LOCATION_CONSTRAINT1 = """
    <rsc_location id="location-constraint1" rsc="R2" node="foo" score="INFINITY"/>
"""
FIXTURE_LOCATION_CONSTRAINT2 = """
    <rsc_location id="location-constraint2" rsc="R2-group" node="foo" score="INFINITY"/>
"""
FIXTURE_LOCATION_CONSTRAINT3 = """
    <rsc_location id="location-constraint3" rsc="R2-clone" node="foo" score="INFINITY"/>
"""
FIXTURE_ALL_LOCATION_CONSTRAINTS = f"""
    {FIXTURE_LOCATION_CONSTRAINT1}
    {FIXTURE_LOCATION_CONSTRAINT2}
    {FIXTURE_LOCATION_CONSTRAINT3}
"""
FIXTURE_SET_CONSTRAINT = """
    <rsc_colocation score="10" id="colocation-constraint">
        <resource_set role="Started" id="set1">
            <resource_ref id="R3"/>
            <resource_ref id="R4"/>
        </resource_set>
    </rsc_colocation>
"""
FIXTURE_ALL_CONSTRAINTS_XML = f"""
    {FIXTURE_ALL_LOCATION_CONSTRAINTS}
    {FIXTURE_SET_CONSTRAINT}
"""


def fixture_message_not_deleting_resources_not_live(resource_ids):
    resources = format_plural(resource_ids, "resource")
    are = format_plural(resource_ids, "is")
    return (
        "Warning: Resources are not going to be stopped before deletion "
        "because the command does not run on a live cluster\n"
        f"Warning: Not checking if {resources} {format_list(resource_ids)} "
        f"{are} stopped before deletion because the command does not run on a "
        "live cluster. Deleting unstopped resources may result in orphaned "
        "resources being present in the cluster.\n"
    )


class ResourceRemoveDeleteBase(
    get_assert_pcs_effect_mixin(
        lambda cib: etree.tostring(etree.parse(cib).find(".//resources"))
    ),
):
    command = ""

    def setUp(self):
        self.temp_cib = get_tmp_file("tier1_resource_remove")
        cib_data = modify_cib_file(
            get_test_resource("cib-empty.xml"),
            resources=f"""
                <resources>
                    {fixture_primitive_xml("R1")}
                    {FIXTURE_GROUP_XML}
                    {FIXTURE_CLONED_GROUP_XML}
                </resources>
            """,
            constraints=f"""
                <constraints>
                    {FIXTURE_ALL_CONSTRAINTS_XML}
                </constraints>
            """,
            tags=f"""
                <tags>
                    {FIXTURE_TAG_XML}
                </tags>
            """,
        )
        write_data_to_tmpfile(cib_data, self.temp_cib)
        self.pcs_runner = PcsRunner(self.temp_cib.name)

    def assert_constraints(self, constraints_xml):
        self.assert_resources_xml_in_cib(
            constraints_xml,
            lambda cib: etree.tostring(etree.parse(cib).find(".//constraints")),
        )

    def assert_tags(self, tags_xml):
        self.assert_resources_xml_in_cib(
            tags_xml,
            lambda cib: etree.tostring(etree.parse(cib).find(".//tags")),
        )

    def tearDown(self):
        self.temp_cib.close()

    def test_no_args(self):
        self.assert_pcs_fail(
            ["resource", self.command],
            stderr_start=f"\nUsage: pcs resource {self.command}...",
        )

    def test_nonexistent_resource(self):
        self.assert_pcs_fail(
            ["resource", self.command, "nonexistent"],
            stderr_full="Error: Unable to find resource: 'nonexistent'\n",
        )

    def test_primitive(self):
        self.assert_effect_single(
            ["resource", self.command, "R1"],
            f"""
            <resources>
                {FIXTURE_GROUP_XML}
                {FIXTURE_CLONED_GROUP_XML}
            </resources>
            """,
            stderr_full=fixture_message_not_deleting_resources_not_live(["R1"]),
        )
        self.assert_constraints(
            f"""
                <constraints>
                    {FIXTURE_ALL_CONSTRAINTS_XML}
                </constraints>
            """
        )
        self.assert_tags(
            f"""
                <tags>
                    {FIXTURE_TAG_XML}
                </tags>
            """
        )

    def test_remove_dependencies(self):
        self.assert_effect_single(
            ["resource", self.command, "R2"],
            f"""
            <resources>
                {fixture_primitive_xml("R1")}
                {FIXTURE_GROUP_XML}
            </resources>
            """,
            stderr_full=fixture_message_not_deleting_resources_not_live(
                ["R2", "R2-clone", "R2-group"]
            )
            + dedent(
                """\
                Removing dependant elements:
                  Clone: 'R2-clone'
                  Group: 'R2-group'
                  Location constraints: 'location-constraint1', 'location-constraint2', 'location-constraint3'
                """
            ),
        )
        self.assert_constraints(
            f"""
            <constraints>
                {FIXTURE_SET_CONSTRAINT}
            </constraints>
            """
        )

    def test_remove_references(self):
        self.assert_effect_single(
            ["resource", self.command, "R3"],
            f"""
            <resources>
                {fixture_primitive_xml("R1")}
                <group id="R3R4-group">
                    {fixture_primitive_xml("R4")}
                </group>
                {FIXTURE_CLONED_GROUP_XML}
            </resources>
            """,
            stderr_full=fixture_message_not_deleting_resources_not_live(["R3"])
            + dedent(
                """\
                Removing references:
                  Resource 'R3' from:
                    Group: 'R3R4-group'
                    Resource set: 'set1'
                    Tag: 'TAG'
                """
            ),
        )
        self.assert_constraints(
            f"""
                <constraints>
                    {FIXTURE_ALL_LOCATION_CONSTRAINTS}
                    <rsc_colocation score="10" id="colocation-constraint">
                        <resource_set role="Started" id="set1">
                            <resource_ref id="R4"/>
                        </resource_set>
                    </rsc_colocation>
                </constraints>
            """
        )
        self.assert_tags(
            """
                <tags>
                    <tag id="TAG">
                        <obj_ref id="R4"/>
                    </tag>
                </tags>
            """
        )

    def test_remove_all_resources(self):
        self.assert_effect_single(
            ["resource", self.command, "R1", "R2", "R3", "R4"],
            "<resources/>",
            stderr_full=fixture_message_not_deleting_resources_not_live(
                ["R1", "R2", "R3", "R4", "R2-clone", "R2-group", "R3R4-group"]
            )
            + dedent(
                """\
                Removing dependant elements:
                  Clone: 'R2-clone'
                  Colocation constraint: 'colocation-constraint'
                  Groups: 'R2-group', 'R3R4-group'
                  Location constraints: 'location-constraint1', 'location-constraint2', 'location-constraint3'
                  Resource set: 'set1'
                  Tag: 'TAG'
                """
            ),
        )
        self.assert_constraints("<constraints/>")
        self.assert_tags("<tags/>")


class ResourceRemove(ResourceRemoveDeleteBase, TestCase):
    command = "remove"


class ResourceDelete(ResourceRemoveDeleteBase, TestCase):
    command = "delete"


class ResourceReferencedInAcl(AssertPcsMixin, TestCase):
    def setUp(self):
        self.temp_cib = get_tmp_file("tier1_resource_remove_referenced_in_acl")
        cib_data = modify_cib_file(
            get_test_resource("cib-empty.xml"),
            resources=f"""
                <resources>
                    {fixture_primitive_xml("R1")}
                </resources>
            """,
            optional_in_conf="""
                <acls>
                    <acl_role id="ROLE">
                        <acl_permission id="PERMISSION" kind="write" reference="R1"/>
                    </acl_role>
                    <acl_target id="TARGET">
                        <role id="ROLE"/>
                    </acl_target>
                </acls>
            """,
        )
        write_data_to_tmpfile(cib_data, self.temp_cib)
        self.pcs_runner = PcsRunner(self.temp_cib.name)

    def tearDown(self):
        self.temp_cib.close()

    def test_remove_primitive(self):
        self.assert_pcs_success(
            ["resource", "delete", "R1"],
            stderr_full=fixture_message_not_deleting_resources_not_live(["R1"])
            + dedent(
                """\
                Removing dependant element:
                  Acl permission: 'PERMISSION'
                Removing references:
                  Acl permission 'PERMISSION' from:
                    Acl role: 'ROLE'
                """
            ),
        )



from typing import Any

from pcs import (
    status,
    usage,
)
from pcs.cli.booth.command import status as booth_status_cmd
from pcs.cli.common.errors import CmdLineInputError
from pcs.cli.common.parse_args import (
    Argv,
    InputModifiers,
)
from pcs.cli.common.routing import create_router
from pcs.cli.query import resource
from pcs.cli.status import command as status_command
from pcs.pcsd import pcsd_status_cmd
from pcs.qdevice import qdevice_status_cmd
from pcs.quorum import quorum_status_cmd
from pcs.resource import resource_status


def _query_resource_router(
    lib: Any, argv: Argv, modifiers: InputModifiers
) -> None:
    if not argv:
        raise CmdLineInputError()

    if len(argv) < 2:
        # show correct usage when resource_id is missing in the command
        argv.insert(0, "")

    # swap resource_id with next keyword in the command to be able to use router
    argv[0], argv[1] = argv[1], argv[0]

    create_router(
        {
            "exists": resource.exists,
            "is-in-bundle": resource.is_in_bundle,
            "is-in-clone": resource.is_in_clone,
            "is-in-group": resource.is_in_group,
            "is-state": resource.is_state,
            "is-stonith": resource.is_stonith,
            "is-type": resource.is_type,
            "get-type": resource.get_type,
            "get-members": resource.get_members,
            "get-nodes": resource.get_nodes,
            "get-index-in-group": resource.get_index_in_group,
        },
        ["status", "query", "resource", "<resource-id>"],
    )(lib, argv, modifiers)


status_cmd = create_router(
    {
        "help": lambda lib, argv, modifiers: print(usage.status(argv)),
        "booth": booth_status_cmd,
        "corosync": status.corosync_status,
        "cluster": status.cluster_status,
        "nodes": status.nodes_status,
        "pcsd": pcsd_status_cmd,
        "qdevice": qdevice_status_cmd,
        "quorum": quorum_status_cmd,
        "resources": resource_status,
        "xml": status.xml_status,
        "status": status.full_status,
        "query": create_router(
            {"resource": _query_resource_router}, ["status", "query"]
        ),
        "wait": status_command.wait_for_pcmk_idle,
    },
    ["status"],
    default_cmd="status",
)



from unittest import TestCase

from lxml import etree

from pcs_test.tools.cib import get_assert_pcs_effect_mixin
from pcs_test.tools.misc import get_test_resource as rc
from pcs_test.tools.misc import (
    get_tmp_file,
    write_data_to_tmpfile,
)
from pcs_test.tools.pcs_runner import PcsRunner
from pcs_test.tools.xml import XmlManipulation


class ResourceStonithIsForbidden(
    TestCase,
    get_assert_pcs_effect_mixin(
        lambda cib: etree.tostring(etree.parse(cib).findall(".//resources")[0])
    ),
):
    # pylint: disable=too-many-public-methods
    empty_cib = rc("cib-empty.xml")

    def setUp(self):
        self.temp_cib = get_tmp_file(
            "tier1_cib_resource_resource_stonith_is_forbidden"
        )
        self.pcs_runner = PcsRunner(self.temp_cib.name)
        xml_manip = XmlManipulation.from_file(self.empty_cib)
        xml_manip.append_to_first_tag_name(
            "resources",
            """
            <primitive id="S1" class="stonith" type="fence_kdump">
                <operations>
                    <op name="monitor" interval="60s"
                        id="S1-monitor-interval-60s"/>
                </operations>
            </primitive>
            """,
        )
        write_data_to_tmpfile(str(xml_manip), self.temp_cib)

    def tearDown(self):
        self.temp_cib.close()

    def test_rsc_config(self):
        self.assert_pcs_fail(
            "resource config S1".split(),
            stderr_full=(
                "Warning: Unable to find resource 'S1'\n"
                "Error: No resource found\n"
            ),
        )

    def test_rsc_create(self):
        self.assert_pcs_fail(
            "resource create S1 stonith:fence_pcsmock_minimal".split(),
            stderr_full=(
                "Error: This command does not accept stonith resource. "
                "Use 'pcs stonith create' command instead.\n"
                "Error: Errors have occurred, therefore pcs is unable to continue\n"
            ),
        )

    def test_rsc_create_in_bundle(self):
        self.assert_pcs_fail(
            "resource create S1 stonith:fence_pcsmock_minimal bundle B".split(),
            stderr_full=(
                "Error: This command does not accept stonith resource. "
                "Use 'pcs stonith create' command instead.\n"
                "Error: Errors have occurred, therefore pcs is unable to continue\n"
            ),
        )

    def test_rsc_create_in_clone(self):
        self.assert_pcs_fail(
            "resource create S1 stonith:fence_pcsmock_minimal clone".split(),
            stderr_full=(
                "Error: This command does not accept stonith resource. "
                "Use 'pcs stonith create' command instead.\n"
                "Error: Errors have occurred, therefore pcs is unable to continue\n"
            ),
        )

    def test_rsc_create_in_group(self):
        self.assert_pcs_fail(
            "resource create S1 stonith:fence_pcsmock_minimal group G --future".split(),
            stderr_full=(
                "Error: This command does not accept stonith resource. "
                "Use 'pcs stonith create' command instead.\n"
                "Error: Errors have occurred, therefore pcs is unable to continue\n"
            ),
        )

    def test_rsc_delete(self):
        self.assert_pcs_fail(
            "resource delete S1".split(),
            stderr_full=(
                "Error: This command cannot remove stonith resource: 'S1'. "
                "Use 'pcs stonith remove' instead.\n"
            ),
        )

    def test_rsc_remove(self):
        self.assert_pcs_fail(
            "resource remove S1".split(),
            stderr_full=(
                "Error: This command cannot remove stonith resource: 'S1'. "
                "Use 'pcs stonith remove' instead.\n"
            ),
        )

    def test_rsc_enable(self):
        self.assert_pcs_fail(
            "resource enable S1".split(),
            stderr_full=(
                "Error: This command does not accept stonith resources. "
                "Please use 'pcs stonith enable' instead.\n"
            ),
        )

    def test_rsc_disable(self):
        self.assert_pcs_fail(
            "resource disable S1".split(),
            stderr_full=(
                "Error: This command does not accept stonith resources. "
                "Please use 'pcs stonith disable' instead.\n"
            ),
        )

    def test_rsc_move_with_constraint(self):
        self.assert_pcs_fail(
            "resource move-with-constraint S1".split(),
            stderr_full=(
                "Error: This command does not accept stonith resource.\n"
                "Error: Errors have occurred, therefore pcs is unable to continue\n"
            ),
        )

    def test_rsc_ban(self):
        self.assert_pcs_fail(
            "resource ban S1".split(),
            stderr_full=(
                "Error: This command does not accept stonith resource.\n"
                "Error: Errors have occurred, therefore pcs is unable to continue\n"
            ),
        )

    def test_rsc_clear(self):
        self.assert_pcs_fail(
            "resource clear S1".split(),
            stderr_full=(
                "Error: This command does not accept stonith resource.\n"
                "Error: Errors have occurred, therefore pcs is unable to continue\n"
            ),
        )

    def test_rsc_update(self):
        self.assert_pcs_fail(
            "resource update S1".split(),
            stderr_full=(
                "Error: This command does not accept stonith resources. "
                "Please use 'pcs stonith update' instead.\n"
            ),
        )

    def test_rsc_op_add(self):
        self.assert_pcs_fail(
            "resource op add S1 monitor timeout=30".split(),
            stderr_full=(
                "Error: This command does not accept stonith resources. "
                "Please use 'pcs stonith op add' instead.\n"
            ),
        )

    def test_rsc_op_delete(self):
        self.assert_pcs_fail(
            "resource op delete S1".split(),
            stderr_full=(
                "Error: This command does not accept stonith resources. "
                "Please use 'pcs stonith op delete' instead.\n"
            ),
        )

    def test_rsc_op_remove(self):
        self.assert_pcs_fail(
            "resource op remove S1".split(),
            stderr_full=(
                "Error: This command does not accept stonith resources. "
                "Please use 'pcs stonith op delete' instead.\n"
            ),
        )

    def test_rsc_meta(self):
        self.assert_pcs_fail(
            "resource meta S1".split(),
            stderr_full=(
                "Error: This command does not accept stonith resources. "
                "Please use 'pcs stonith meta' instead.\n"
            ),
        )

    def test_rsc_group_add(self):
        self.assert_pcs_fail(
            "resource group add G S1".split(),
            stderr_full=(
                "Error: 'S1' is a stonith resource, stonith resources cannot be put into a group\n"
                "Error: Errors have occurred, therefore pcs is unable to continue\n"
            ),
        )

    def test_rsc_clone(self):
        self.assert_pcs_fail(
            "resource clone S1".split(),
            stderr_full=(
                "Error: This command does not accept stonith resources.\n"
            ),
        )

    def test_rsc_promotable(self):
        self.assert_pcs_fail(
            "resource promotable S1".split(),
            stderr_full=(
                "Error: This command does not accept stonith resources.\n"
            ),
        )

    def test_rsc_manage(self):
        self.assert_pcs_fail(
            "resource manage S1".split(),
            stderr_full=(
                "Error: This command does not accept stonith resources.\n"
            ),
        )

    def test_rsc_unmanage(self):
        self.assert_pcs_fail(
            "resource unmanage S1".split(),
            stderr_full=(
                "Error: This command does not accept stonith resources.\n"
            ),
        )

    def test_rsc_relocate_dry_run(self):
        self.assert_pcs_fail(
            "resource relocate dry-run S1".split(),
            stderr_full=(
                "Error: This command does not accept stonith resources.\n"
            ),
        )

    def test_rsc_utilization(self):
        self.assert_pcs_fail(
            "resource utilization S1 test1=10".split(),
            stderr_full=(
                "Error: This command does not accept stonith resources.\n"
            ),
        )

    def test_rsc_relations(self):
        self.assert_pcs_fail(
            "resource relations S1".split(),
            stderr_full=(
                "Error: This command does not accept stonith resource.\n"
                "Error: Errors have occurred, therefore pcs is unable to continue\n"
            ),
        )



from shlex import quote
from typing import Iterable

from pcs.cli.common.output import (
    INDENT_STEP,
    pairs_to_cmd,
)
from pcs.common.pacemaker.constraint import (
    CibConstraintTicketAttributesDto,
    CibConstraintTicketDto,
    CibConstraintTicketSetDto,
)
from pcs.common.str_tools import (
    format_optional,
    indent,
    pairs_to_text,
)

from . import set as _set


def _attributes_to_pairs(
    attributes_dto: CibConstraintTicketAttributesDto,
) -> list[tuple[str, str]]:
    pairs = []
    if attributes_dto.loss_policy:
        pairs.append(("loss-policy", str(attributes_dto.loss_policy)))
    return pairs


def plain_constraint_to_text(
    constraint_dto: CibConstraintTicketDto,
    with_id: bool,
) -> list[str]:
    result = [
        "{role}resource '{resource}' depends on ticket '{ticket}'".format(
            role=format_optional(constraint_dto.role),
            resource=constraint_dto.resource_id,
            ticket=constraint_dto.attributes.ticket,
        )
    ]
    if with_id:
        result[0] += f" (id: {constraint_dto.attributes.constraint_id})"
    result.extend(
        indent(
            pairs_to_text(_attributes_to_pairs(constraint_dto.attributes)),
            indent_step=INDENT_STEP,
        )
    )
    return result


def set_constraint_to_text(
    constraint_dto: CibConstraintTicketSetDto,
    with_id: bool,
) -> list[str]:
    return _set.set_constraint_to_text(
        constraint_dto.attributes.constraint_id,
        pairs_to_text(
            [("ticket", constraint_dto.attributes.ticket)]
            + _attributes_to_pairs(constraint_dto.attributes)
        ),
        constraint_dto.resource_sets,
        with_id,
    )


def constraints_to_text(
    plain_dtos: Iterable[CibConstraintTicketDto],
    set_dtos: Iterable[CibConstraintTicketSetDto],
    with_id: bool,
) -> list[str]:
    result = []
    if plain_dtos:
        result.append("Ticket Constraints:")
        for constraint_dto in plain_dtos:
            result.extend(
                indent(
                    plain_constraint_to_text(constraint_dto, with_id),
                    indent_step=INDENT_STEP,
                )
            )
    if set_dtos:
        result.append("Ticket Set Constraints:")
        for set_constraint_dto in set_dtos:
            result.extend(
                indent(
                    set_constraint_to_text(set_constraint_dto, with_id),
                    indent_step=INDENT_STEP,
                )
            )
    return result


def _attributes_to_cmd_pairs(
    attributes_dto: CibConstraintTicketAttributesDto,
) -> list[tuple[str, str]]:
    return [("id", attributes_dto.constraint_id)] + _attributes_to_pairs(
        attributes_dto
    )


def plain_constraint_to_cmd(
    constraint_dto: CibConstraintTicketDto,
) -> list[str]:
    result = [
        "pcs -- constraint ticket add {ticket} {role}{resource_id}".format(
            ticket=quote(constraint_dto.attributes.ticket),
            role=format_optional(constraint_dto.role),
            resource_id=quote(constraint_dto.resource_id),
        )
    ]
    params = pairs_to_cmd(_attributes_to_cmd_pairs(constraint_dto.attributes))
    if params:
        result.extend(indent([params], indent_step=INDENT_STEP))
    return result


def set_constraint_to_cmd(
    constraint_dto: CibConstraintTicketSetDto,
) -> list[str]:
    result = ["pcs -- constraint ticket"]
    for resource_set in constraint_dto.resource_sets:
        set_cmd_part = _set.resource_set_to_cmd(resource_set)
        if not set_cmd_part:
            return []
        result.extend(indent(set_cmd_part, indent_step=INDENT_STEP))
    params = pairs_to_cmd(
        _attributes_to_cmd_pairs(constraint_dto.attributes)
        + [("ticket", constraint_dto.attributes.ticket)]
    )
    if params:
        result.extend(indent([f"setoptions {params}"], indent_step=INDENT_STEP))
    return result



from typing import Any

from pcs.cli.common.errors import CmdLineInputError
from pcs.cli.common.parse_args import (
    Argv,
    InputModifiers,
)


def wait_for_pcmk_idle(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    """
    Options: no options
    """
    modifiers.ensure_only_supported()

    if len(argv) > 1:
        raise CmdLineInputError()

    lib.cluster.wait_for_pcmk_idle(argv[0] if argv else None)



from unittest import mock

from pcs.cli.stonith import command

from pcs_test.tier0.cli.resource.test_update import ResourceMeta
from pcs_test.tools.misc import dict_to_modifiers

RESOURCE_ID = "resource-id"


class StonithMeta(ResourceMeta):
    command = staticmethod(command.meta)

    def setUp(self):
        super().setUp()
        self.lib.resource.is_any_resource_except_stonith = mock.Mock()
        self.lib.resource.is_any_resource_except_stonith.return_value = False

        mock_get_resource_status_msg_patcher = mock.patch(
            "pcs.cli.stonith.command.get_resource_status_msg"
        )
        self.addCleanup(mock_get_resource_status_msg_patcher.stop)
        self.mock_get_resource_status_msg = (
            mock_get_resource_status_msg_patcher.start()
        )

        mock_deprecation_warning_patcher = mock.patch(
            "pcs.cli.stonith.command.deprecation_warning"
        )
        self.addCleanup(mock_deprecation_warning_patcher.stop)
        self.mock_deprecation_warning = mock_deprecation_warning_patcher.start()

    def _assert_no_wait_or_stonith(self):
        self.lib.resource.is_any_resource_except_stonith.assert_called_once_with(
            [RESOURCE_ID]
        )
        self.wait_for_pcmk_idle.assert_not_called()
        self.mock_get_resource_status_msg.assert_not_called()

    @mock.patch("pcs.cli.reports.output.print_to_stderr")
    def test_stonith_forbidden(self, mock_print):
        self.lib.resource.is_any_resource_except_stonith.return_value = True
        self.assertRaises(
            SystemExit,
            lambda: self.command(
                self.lib,
                [RESOURCE_ID, "meta-attr1=value1"],
                dict_to_modifiers({}),
            ),
        )
        self.update_meta.assert_not_called()
        mock_print.assert_called_once_with(
            "Error: This command does not accept resources. Please use 'pcs "
            "resource meta' instead."
        )
        self._assert_no_wait_or_stonith()

    @mock.patch("pcs.cli.stonith.command.print")
    def test_with_wait_zero(self, _):
        self.command(
            self.lib,
            [RESOURCE_ID, "meta-attr1=value1"],
            dict_to_modifiers({"wait": "0"}),
        )
        self.update_meta.assert_called_once_with(
            RESOURCE_ID, {"meta-attr1": "value1"}, []
        )
        self.lib.resource.is_any_resource_except_stonith.assert_called_once_with(
            [RESOURCE_ID]
        )
        self.wait_for_pcmk_idle.assert_called_once_with(0)



from functools import partial
from unittest import TestCase, mock

from pcs import settings
from pcs.common import reports
from pcs.common.host import Destination
from pcs.common.reports import codes as report_codes
from pcs.lib.commands.remote_node import (
    node_remove_remote as node_remove_remote_orig,
)

from pcs_test.tier0.lib.commands.remote_node.fixtures_add import (
    FAIL_HTTP_KWARGS,
    report_manage_services_connection_failed,
)
from pcs_test.tier0.lib.commands.remote_node.fixtures_remove import (
    EnvConfigMixin,
    base_reports_for_host,
    report_authkey_remove_failed,
    report_pcmk_remote_disable_failed,
    report_pcmk_remote_stop_failed,
    report_remove_file_connection_failed,
)
from pcs_test.tools import fixture
from pcs_test.tools.command_env import get_env_tools
from pcs_test.tools.misc import get_test_resource as rc

NODE_NAME = "node-name"
NODE_DEST_LIST = [Destination("node-addr", 2224)]
REMOTE_HOST = "remote-host"
REMOTE_DEST_LIST = [Destination("remote-addr", 2224)]
NODE_1 = "node-1"
NODE_2 = "node-2"


def node_remove_remote(env, *args, node_identifier=REMOTE_HOST, **kwargs):
    node_remove_remote_orig(env, node_identifier, *args, **kwargs)


FIXTURE_RESOURCES_CIB = """
    <resources>
        <primitive class="ocf" id="{0}" provider="pacemaker" type="remote">
            <instance_attributes id="node-name-instance_attributes">
                <nvpair
                    id="node-name-instance_attributes-server"
                    name="server" value="{1}"
                />
            </instance_attributes>
        </primitive>
    </resources>
""".format(
    NODE_NAME,
    REMOTE_HOST,
)

FIXTURE_RESOURCES_STATE_STARTED = f"""
    <resources>
        <resource id="{NODE_NAME}" managed="true" role="Started"/>
    </resources>
"""

FIXTURE_RESOURCES_STATE_STOPPED = f"""
    <resources>
        <resource id="{NODE_NAME}" managed="true" role="Stopped"/>
    </resources>
"""


REPORTS = fixture.ReportSequenceBuilder().info(
    report_codes.CIB_REMOVE_RESOURCES,
    id_list=[NODE_NAME],
    _name="cib_remove_resources",
).fixtures + base_reports_for_host(NODE_NAME)


get_env_tools = partial(
    get_env_tools, local_extensions={"local": EnvConfigMixin}
)


@mock.patch.object(
    settings, "pacemaker_api_result_schema", rc("pcmk_rng/api/api-result.rng")
)
class RemoveRemote(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)
        self.config.env.set_known_hosts_dests(
            {
                NODE_NAME: NODE_DEST_LIST,
            }
        )

    def find_by(self, identifier):
        self.config.runner.cib.load(resources=FIXTURE_RESOURCES_CIB)
        self.config.runner.pcmk.load_state(
            resources=FIXTURE_RESOURCES_STATE_STOPPED
        )
        self.config.local.destroy_pacemaker_remote(
            label=NODE_NAME, dest_list=NODE_DEST_LIST
        )
        self.config.local.remove_authkey(
            communication_list=[
                dict(label=NODE_NAME, dest_list=NODE_DEST_LIST)
            ],
        )
        self.config.env.push_cib(resources="<resources/>")
        self.config.runner.pcmk.remove_node(NODE_NAME)

        node_remove_remote(
            self.env_assist.get_env(), node_identifier=identifier
        )
        self.env_assist.assert_reports(REPORTS)

    def test_success_base(self):
        self.find_by(REMOTE_HOST)

    def test_can_find_by_node_name(self):
        self.find_by(NODE_NAME)


@mock.patch.object(
    settings, "pacemaker_api_result_schema", rc("pcmk_rng/api/api-result.rng")
)
class RemoveRemoteOthers(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)
        self.config.env.set_known_hosts_dests(
            {
                NODE_NAME: NODE_DEST_LIST,
            }
        )

    def test_can_skip_all_offline(self):
        self.config.runner.cib.load(resources=FIXTURE_RESOURCES_CIB)
        self.config.runner.pcmk.load_state(
            resources=FIXTURE_RESOURCES_STATE_STOPPED
        )
        self.config.local.destroy_pacemaker_remote(
            label=NODE_NAME, dest_list=NODE_DEST_LIST, **FAIL_HTTP_KWARGS
        )
        self.config.local.remove_authkey(
            communication_list=[
                dict(label=NODE_NAME, dest_list=NODE_DEST_LIST)
            ],
            **FAIL_HTTP_KWARGS,
        )
        self.config.env.push_cib(resources="<resources/>")
        self.config.runner.pcmk.remove_node(NODE_NAME)

        node_remove_remote(
            self.env_assist.get_env(), [reports.codes.SKIP_OFFLINE_NODES]
        )
        my_reports = REPORTS.copy()
        my_reports.replace(
            "pcmk_remote_disable_success",
            report_manage_services_connection_failed(NODE_NAME).to_warn(),
        )
        my_reports.remove("pcmk_remote_stop_success")
        my_reports.replace(
            "authkey_remove_success",
            report_remove_file_connection_failed(NODE_NAME).to_warn(),
        )
        self.env_assist.assert_reports(my_reports)

    def test_fail_when_identifier_not_found(self):
        self.config.runner.cib.load(resources=FIXTURE_RESOURCES_CIB)
        self.env_assist.assert_raise_library_error(
            lambda: node_remove_remote(
                self.env_assist.get_env(), node_identifier="NONEXISTENT"
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    report_codes.NODE_NOT_FOUND,
                    node="NONEXISTENT",
                    searched_types=["remote"],
                )
            ]
        )

    def test_fails_when_some_started(self):
        self.config.runner.cib.load(resources=FIXTURE_RESOURCES_CIB)
        self.config.runner.pcmk.load_state(
            resources=FIXTURE_RESOURCES_STATE_STARTED
        )
        self.env_assist.assert_raise_library_error(
            lambda: node_remove_remote(
                self.env_assist.get_env(), node_identifier=REMOTE_HOST
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    report_codes.CANNOT_REMOVE_RESOURCES_NOT_STOPPED,
                    resource_id_list=[NODE_NAME],
                    force_code=report_codes.FORCE,
                )
            ]
        )

    def test_some_started_force(self):
        self.config.runner.cib.load(resources=FIXTURE_RESOURCES_CIB)
        self.config.runner.pcmk.load_state(
            resources=FIXTURE_RESOURCES_STATE_STARTED
        )
        self.config.local.destroy_pacemaker_remote(
            label=NODE_NAME, dest_list=NODE_DEST_LIST
        )
        self.config.local.remove_authkey(
            communication_list=[
                dict(label=NODE_NAME, dest_list=NODE_DEST_LIST)
            ],
        )
        self.config.env.push_cib(resources="<resources/>")
        self.config.runner.pcmk.remove_node(NODE_NAME)
        node_remove_remote(
            self.env_assist.get_env(),
            node_identifier=REMOTE_HOST,
            force_flags=[reports.codes.FORCE],
        )

        self.env_assist.assert_reports(
            fixture.ReportSequenceBuilder()
            .warn(
                reports.codes.CANNOT_REMOVE_RESOURCES_NOT_STOPPED,
                resource_id_list=[NODE_NAME],
            )
            .fixtures
            + REPORTS
        )


@mock.patch.object(
    settings, "pacemaker_api_result_schema", rc("pcmk_rng/api/api-result.rng")
)
class MultipleResults(TestCase):
    fixture_multi_resources = """
        <resources>
            <primitive class="ocf" id="{0}" provider="pacemaker" type="remote">
                <instance_attributes id="node-name-instance_attributes">
                    <nvpair
                        id="node-name-instance_attributes-server"
                        name="server" value="{1}"
                    />
                </instance_attributes>
            </primitive>
            <primitive class="ocf" id="{1}" provider="pacemaker" type="remote">
                <instance_attributes id="node-name-instance_attributes">
                    <nvpair
                        id="node-name-instance_attributes-server"
                        name="server" value="{2}"
                    />
                </instance_attributes>
            </primitive>
        </resources>
    """.format(NODE_NAME, REMOTE_HOST, "OTHER-REMOTE")

    fixture_multi_resources_state = f"""
        <resources>
            <resource id="{NODE_NAME}" managed="true" role="Stopped"/>
            <resource id="{REMOTE_HOST}" managed="true" role="Stopped"/>
        </resources>
    """

    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)
        self.config.runner.cib.load(resources=self.fixture_multi_resources)
        self.report_multiple_results = fixture.error(
            report_codes.MULTIPLE_RESULTS_FOUND,
            force_code=report_codes.FORCE,
            result_identifier_list=[
                NODE_NAME,
                REMOTE_HOST,
            ],
            result_type="resource",
            search_description=REMOTE_HOST,
        )
        self.config.env.set_known_hosts_dests(
            {
                NODE_NAME: NODE_DEST_LIST,
                REMOTE_HOST: REMOTE_DEST_LIST,
            }
        )

    def test_fail(self):
        self.env_assist.assert_raise_library_error(
            lambda: node_remove_remote(
                self.env_assist.get_env(), node_identifier=REMOTE_HOST
            )
        )
        self.env_assist.assert_reports([self.report_multiple_results])

    def test_force(self):
        self.config.runner.pcmk.load_state(
            resources=self.fixture_multi_resources_state
        )
        self.config.local.destroy_pacemaker_remote(
            communication_list=[
                dict(label=NODE_NAME, dest_list=NODE_DEST_LIST),
                dict(label=REMOTE_HOST, dest_list=REMOTE_DEST_LIST),
            ]
        )
        self.config.local.remove_authkey(
            communication_list=[
                dict(label=NODE_NAME, dest_list=NODE_DEST_LIST),
                dict(label=REMOTE_HOST, dest_list=REMOTE_DEST_LIST),
            ],
        )
        self.config.env.push_cib(resources="<resources/>")
        self.config.runner.pcmk.remove_node(
            NODE_NAME, name="remove_node.node_name"
        )
        self.config.runner.pcmk.remove_node(
            REMOTE_HOST, name="remove_node.remote_host"
        )

        node_remove_remote(
            self.env_assist.get_env(),
            node_identifier=REMOTE_HOST,
            force_flags=[reports.codes.FORCE],
        )
        my_reports = REPORTS.copy()
        my_reports.replace(
            "cib_remove_resources",
            REPORTS["cib_remove_resources"].adapt(
                id_list=[NODE_NAME, REMOTE_HOST]
            ),
        )
        my_reports.replace(
            "pcmk_remote_disable_stop_started",
            REPORTS["pcmk_remote_disable_stop_started"].adapt(
                node_list=[NODE_NAME, REMOTE_HOST]
            ),
        )
        my_reports.append(
            REPORTS["pcmk_remote_disable_success"].adapt(node=REMOTE_HOST)
        )
        my_reports.append(
            REPORTS["pcmk_remote_stop_success"].adapt(node=REMOTE_HOST)
        )
        my_reports.replace(
            "authkey_remove_started",
            REPORTS["authkey_remove_started"].adapt(
                node_list=[NODE_NAME, REMOTE_HOST]
            ),
        )
        my_reports.append(
            REPORTS["authkey_remove_success"].adapt(node=REMOTE_HOST)
        )
        my_reports.append(self.report_multiple_results.to_warn())
        self.env_assist.assert_reports(my_reports)


class AuthkeyRemove(TestCase):
    def setUp(self):
        rng_path_patcher = mock.patch.object(
            settings,
            "pacemaker_api_result_schema",
            rc("pcmk_rng/api/api-result.rng"),
        )
        self.addCleanup(rng_path_patcher.stop)
        rng_path_patcher.start()

        self.env_assist, self.config = get_env_tools(self)
        self.config.env.set_known_hosts_dests(
            {
                NODE_NAME: NODE_DEST_LIST,
            }
        )
        self.config.runner.cib.load(resources=FIXTURE_RESOURCES_CIB)
        self.config.runner.pcmk.load_state(
            resources=FIXTURE_RESOURCES_STATE_STOPPED
        )

    def test_fails_when_offline(self):
        self.config.local.destroy_pacemaker_remote(
            label=NODE_NAME, dest_list=NODE_DEST_LIST
        )
        self.config.local.remove_authkey(
            communication_list=[
                dict(label=NODE_NAME, dest_list=NODE_DEST_LIST)
            ],
            **FAIL_HTTP_KWARGS,
        )
        self.env_assist.assert_raise_library_error(
            lambda: node_remove_remote(self.env_assist.get_env())
        )
        my_reports = REPORTS.copy()
        my_reports.remove("cib_remove_resources")
        my_reports.replace(
            "authkey_remove_success",
            report_remove_file_connection_failed(NODE_NAME),
        )
        self.env_assist.assert_reports(my_reports)

    def test_fails_when_remotely_fails(self):
        self.config.local.destroy_pacemaker_remote(
            label=NODE_NAME, dest_list=NODE_DEST_LIST
        )
        self.config.local.remove_authkey(
            communication_list=[
                dict(label=NODE_NAME, dest_list=NODE_DEST_LIST)
            ],
            result={
                "code": "unexpected",
                "message": "Access denied",
            },
        )
        self.env_assist.assert_raise_library_error(
            lambda: node_remove_remote(self.env_assist.get_env())
        )
        my_reports = REPORTS.copy()
        my_reports.remove("cib_remove_resources")
        my_reports.replace(
            "authkey_remove_success",
            report_authkey_remove_failed(NODE_NAME),
        )
        self.env_assist.assert_reports(my_reports)

    def test_forceable_when_remotely_fail(self):
        self.config.local.destroy_pacemaker_remote(
            label=NODE_NAME, dest_list=NODE_DEST_LIST
        )
        self.config.local.remove_authkey(
            communication_list=[
                dict(label=NODE_NAME, dest_list=NODE_DEST_LIST)
            ],
            result={
                "code": "unexpected",
                "message": "Access denied",
            },
        )
        self.config.env.push_cib(resources="<resources/>")
        self.config.runner.pcmk.remove_node(NODE_NAME)

        node_remove_remote(self.env_assist.get_env(), [reports.codes.FORCE])
        my_reports = REPORTS.copy()
        my_reports.replace(
            "authkey_remove_success",
            report_authkey_remove_failed(NODE_NAME).to_warn(),
        )
        self.env_assist.assert_reports(my_reports)


class PcmkRemoteServiceDestroy(TestCase):
    def setUp(self):
        rng_path_patcher = mock.patch.object(
            settings,
            "pacemaker_api_result_schema",
            rc("pcmk_rng/api/api-result.rng"),
        )
        self.addCleanup(rng_path_patcher.stop)
        rng_path_patcher.start()

        self.env_assist, self.config = get_env_tools(self)
        self.config.runner.cib.load(resources=FIXTURE_RESOURCES_CIB)
        self.config.runner.pcmk.load_state(
            resources=FIXTURE_RESOURCES_STATE_STOPPED
        )
        self.config.env.set_known_hosts_dests(
            {
                NODE_NAME: NODE_DEST_LIST,
            }
        )

    def test_fails_when_offline(self):
        self.config.local.destroy_pacemaker_remote(
            label=NODE_NAME, dest_list=NODE_DEST_LIST, **FAIL_HTTP_KWARGS
        )
        self.env_assist.assert_raise_library_error(
            lambda: node_remove_remote(self.env_assist.get_env())
        )
        my_reports = REPORTS[:"pcmk_remote_disable_success"]
        my_reports.remove("cib_remove_resources")
        my_reports.append(report_manage_services_connection_failed(NODE_NAME))
        self.env_assist.assert_reports(my_reports)

    def test_fails_when_remotely_fails(self):
        self.config.local.destroy_pacemaker_remote(
            label=NODE_NAME,
            dest_list=NODE_DEST_LIST,
            result={
                "code": "fail",
                "message": "Action failed",
            },
        )
        self.env_assist.assert_raise_library_error(
            lambda: node_remove_remote(self.env_assist.get_env())
        )
        my_reports = REPORTS[:"pcmk_remote_disable_success"]
        my_reports.remove("cib_remove_resources")
        my_reports.append(report_pcmk_remote_disable_failed(NODE_NAME))
        my_reports.append(report_pcmk_remote_stop_failed(NODE_NAME))
        self.env_assist.assert_reports(my_reports)

    def test_forceable_when_remotely_fail(self):
        self.config.local.destroy_pacemaker_remote(
            label=NODE_NAME,
            dest_list=NODE_DEST_LIST,
            result={
                "code": "fail",
                "message": "Action failed",
            },
        ).local.remove_authkey(
            communication_list=[
                dict(label=NODE_NAME, dest_list=NODE_DEST_LIST)
            ],
        )
        self.config.env.push_cib(resources="<resources/>")
        self.config.runner.pcmk.remove_node(NODE_NAME)

        node_remove_remote(self.env_assist.get_env(), [reports.codes.FORCE])
        my_reports = REPORTS.copy()
        my_reports.replace(
            "pcmk_remote_disable_success",
            report_pcmk_remote_disable_failed(NODE_NAME).to_warn(),
        )
        my_reports.replace(
            "pcmk_remote_stop_success",
            report_pcmk_remote_stop_failed(NODE_NAME).to_warn(),
        )
        self.env_assist.assert_reports(my_reports)



import shlex
from typing import Optional, Sequence, Union

from pcs.cli.common.output import (
    INDENT_STEP,
    pairs_to_cmd,
)
from pcs.cli.nvset import nvset_dto_to_lines
from pcs.common.pacemaker.alert import (
    CibAlertDto,
    CibAlertListDto,
    CibAlertRecipientDto,
    CibAlertSelectDto,
)
from pcs.common.pacemaker.nvset import CibNvsetDto
from pcs.common.str_tools import (
    format_list,
    format_optional,
    indent,
)


def _description_to_lines(desc: Optional[str]) -> list[str]:
    return [f"Description: {desc}"] if desc else []


def _nvsets_to_lines(label: str, nvsets: Sequence[CibNvsetDto]) -> list[str]:
    if nvsets and nvsets[0].nvpairs:
        return nvset_dto_to_lines(nvset=nvsets[0], nvset_label=label)
    return []


def _recipient_dto_to_lines(recipient_dto: CibAlertRecipientDto) -> list[str]:
    lines = (
        _description_to_lines(recipient_dto.description)
        + [f"Value: {recipient_dto.value}"]
        + _nvsets_to_lines("Attributes", recipient_dto.instance_attributes)
        + _nvsets_to_lines("Meta Attributes", recipient_dto.meta_attributes)
    )
    return [f"Recipient: {recipient_dto.id}"] + indent(
        lines, indent_step=INDENT_STEP
    )


def _recipients_to_lines(
    recipient_dto_list: Sequence[CibAlertRecipientDto],
) -> list[str]:
    if not recipient_dto_list:
        return []
    lines = []
    for recipient_dto in recipient_dto_list:
        lines.extend(_recipient_dto_to_lines(recipient_dto))
    return ["Recipients:"] + indent(lines, indent_step=INDENT_STEP)


def _select_dto_to_lines(select_dto: Optional[CibAlertSelectDto]) -> list[str]:
    if not select_dto:
        return []
    lines = []
    if select_dto.nodes:
        lines.append("nodes")
    if select_dto.fencing:
        lines.append("fencing")
    if select_dto.resources:
        lines.append("resources")
    if select_dto.attributes:
        attr_names = format_list(
            attr.name for attr in select_dto.attributes_select
        )
        lines.append("attributes" + format_optional(attr_names, ": {}"))
    return ["Receives:"] + indent(lines, indent_step=INDENT_STEP)


def alert_dto_to_lines(alert_dto: CibAlertDto) -> list[str]:
    lines = (
        _description_to_lines(alert_dto.description)
        + [f"Path: {alert_dto.path}"]
        + _recipients_to_lines(alert_dto.recipients)
        + _select_dto_to_lines(alert_dto.select)
        + _nvsets_to_lines("Attributes", alert_dto.instance_attributes)
        + _nvsets_to_lines("Meta Attributes", alert_dto.meta_attributes)
    )
    return [f"Alert: {alert_dto.id}"] + indent(lines, indent_step=INDENT_STEP)


def config_dto_to_lines(config_dto: CibAlertListDto) -> list[str]:
    result = []
    for alert_dto in config_dto.alerts:
        result.extend(alert_dto_to_lines(alert_dto))
    return result


def config_dto_to_cmd(config_dto: CibAlertListDto) -> list[str]:
    commands = []
    for alert_dto in config_dto.alerts:
        # alert
        alert_parts = [
            "pcs -- alert create path={path} id={id}".format(
                path=shlex.quote(alert_dto.path), id=shlex.quote(alert_dto.id)
            )
        ] + _desc_instance_meta_to_cmd(alert_dto)
        # TODO export select, once it is supported by pcs
        commands.append(" ".join(alert_parts))
        # recipients
        for recipient_dto in alert_dto.recipients:
            recipient_parts = [
                "pcs -- alert recipient add {alert_id} value={value} id={id}".format(
                    alert_id=shlex.quote(alert_dto.id),
                    value=shlex.quote(recipient_dto.value),
                    id=shlex.quote(recipient_dto.id),
                )
            ] + _desc_instance_meta_to_cmd(recipient_dto)
            commands.append(" ".join(recipient_parts))
    return commands


def _nvset_to_cmd(
    label: Optional[str],
    nvsets: Sequence[CibNvsetDto],
) -> list[str]:
    if nvsets and nvsets[0].nvpairs:
        options = pairs_to_cmd(
            (nvpair.name, nvpair.value) for nvpair in nvsets[0].nvpairs
        )
        if label:
            options = f"{label} {options}"
        return [options]
    return []


def _desc_instance_meta_to_cmd(
    dto: Union[CibAlertDto, CibAlertRecipientDto],
) -> list[str]:
    parts = []
    if dto.description:
        parts.append(
            "description={desc}".format(desc=shlex.quote(dto.description))
        )
    parts.extend(_nvset_to_cmd("options", dto.instance_attributes))
    parts.extend(_nvset_to_cmd("meta", dto.meta_attributes))
    return parts



import json
from typing import Optional
from unittest import TestCase

import pycurl

from pcs.common import file_type_codes, reports
from pcs.common.communication.const import (
    COM_STATUS_ERROR,
    COM_STATUS_SUCCESS,
    COM_STATUS_UNKNOWN_CMD,
)
from pcs.common.communication.dto import InternalCommunicationResultDto
from pcs.common.communication.types import CommunicationResultStatus
from pcs.common.interface.dto import to_dict
from pcs.common.node_communicator import (
    Request,
    RequestData,
    RequestTarget,
    Response,
)
from pcs.common.pcs_cfgsync_dto import SyncConfigsDto
from pcs.lib.communication.pcs_cfgsync import (
    ConfigInfo,
    GetConfigs,
)

from pcs_test.tools import fixture
from pcs_test.tools.custom_mock import (
    MockCurlSimple,
    MockLibraryReportProcessor,
)


def fixture_communication_result_string(
    status: CommunicationResultStatus = COM_STATUS_SUCCESS,
    status_msg: Optional[str] = None,
    report_list: Optional[reports.dto.ReportItemDto] = None,
    data="",
) -> str:
    return json.dumps(
        to_dict(
            InternalCommunicationResultDto(
                status=status,
                status_msg=status_msg,
                report_list=report_list or [],
                data=data,
            )
        )
    )


def fixture_fallback_request(node_label="NODE"):
    return Request(
        RequestTarget(node_label),
        RequestData(
            action="remote/get_configs",
            structured_data=[("cluster_name", "test")],
        ),
    )


def fixture_request_apiv1(node_label="NODE"):
    return Request(
        RequestTarget(node_label),
        RequestData(
            "api/v1/cfgsync-get-configs/v1",
            data=json.dumps({"cluster_name": "test"}),
        ),
    )


def fixture_response(response_code=200, output="", request=None):
    return Response(
        MockCurlSimple(
            info={pycurl.RESPONSE_CODE: response_code},
            output=output,
            request=request,
        ),
        was_connected=True,
    )


def fixture_fallback_response(
    response_code: int = 200, output="", node_label="NODE"
):
    return fixture_response(
        response_code, output, request=fixture_fallback_request(node_label)
    )


def fixture_apiv1_response(
    response_code: int = 200,
    com_status=COM_STATUS_SUCCESS,
    status_msg=None,
    data="",
    node_label="NODE",
    report_list=None,
):
    return fixture_response(
        response_code,
        output=fixture_communication_result_string(
            status=com_status,
            status_msg=status_msg,
            data=data,
            report_list=report_list or [],
        ),
        request=fixture_request_apiv1(node_label),
    )


class GetConfigsApiv1ResponseProcessing(TestCase):
    def setUp(self):
        self.reporter = MockLibraryReportProcessor()
        self.cmd = GetConfigs(self.reporter, "test")

    def assert_response(self, real, expected):
        self.assertEqual(real.action, expected.action)
        self.assertEqual(real.target.label, expected.target.label)
        self.assertEqual(real.target.dest_list, expected.target.dest_list)
        self.assertEqual(real.data, expected.data)

    def test_empty_response(self):
        requests = self.cmd.on_response(fixture_apiv1_response())
        result = self.cmd.on_complete()

        self.assertEqual(requests, [])
        self.assertFalse(result.was_successful)
        self.assertEqual(result.config_files, {})
        self.reporter.assert_reports(
            [fixture.error(reports.codes.INVALID_RESPONSE_FORMAT, node="NODE")]
        )

    def test_404_returns_fallback_request(self):
        requests = self.cmd.on_response(
            fixture_apiv1_response(response_code=404)
        )

        self.assertEqual(len(requests), 1)
        self.assert_response(requests[0], fixture_fallback_request())
        self.reporter.assert_reports([])

    def test_unknown_cmd_fallback_request(self):
        requests = self.cmd.on_response(
            fixture_apiv1_response(com_status=COM_STATUS_UNKNOWN_CMD)
        )

        self.assertEqual(len(requests), 1)
        self.assert_response(requests[0], fixture_fallback_request())
        self.reporter.assert_reports([])

    def test_single_response_with_files(self):
        requests = self.cmd.on_response(
            fixture_apiv1_response(
                data=SyncConfigsDto(
                    cluster_name="test",
                    configs={
                        file_type_codes.PCS_KNOWN_HOSTS: "foo",
                        file_type_codes.PCS_SETTINGS_CONF: "bar",
                    },
                )
            )
        )
        result = self.cmd.on_complete()

        self.assertEqual(requests, [])
        self.assertFalse(result.was_successful)
        self.assertEqual(
            result.config_files,
            {
                file_type_codes.PCS_KNOWN_HOSTS: [ConfigInfo("NODE", "foo")],
                file_type_codes.PCS_SETTINGS_CONF: [ConfigInfo("NODE", "bar")],
            },
        )
        self.reporter.assert_reports([])

    def test_multiple_responses_with_files(self):
        requests = self.cmd.on_response(
            fixture_apiv1_response(
                node_label="NODE-1",
                data=SyncConfigsDto(
                    cluster_name="test",
                    configs={
                        file_type_codes.PCS_KNOWN_HOSTS: "foo-1",
                        file_type_codes.PCS_SETTINGS_CONF: "bar-1",
                    },
                ),
            )
        )
        requests.extend(
            self.cmd.on_response(
                fixture_apiv1_response(
                    node_label="NODE-2",
                    data=SyncConfigsDto(
                        cluster_name="test",
                        configs={
                            file_type_codes.PCS_KNOWN_HOSTS: "foo-2",
                            file_type_codes.PCS_SETTINGS_CONF: "bar-2",
                        },
                    ),
                )
            )
        )
        result = self.cmd.on_complete()

        self.assertEqual(requests, [])
        self.assertTrue(result.was_successful)
        self.assertEqual(
            result.config_files,
            {
                file_type_codes.PCS_KNOWN_HOSTS: [
                    ConfigInfo("NODE-1", "foo-1"),
                    ConfigInfo("NODE-2", "foo-2"),
                ],
                file_type_codes.PCS_SETTINGS_CONF: [
                    ConfigInfo("NODE-1", "bar-1"),
                    ConfigInfo("NODE-2", "bar-2"),
                ],
            },
        )
        self.reporter.assert_reports([])

    def test_invalid_apiv1_format(self):
        requests = self.cmd.on_response(
            fixture_response(output="invalid", request=fixture_request_apiv1())
        )
        result = self.cmd.on_complete()

        self.assertEqual(requests, [])
        self.assertFalse(result.was_successful)
        self.assertEqual(result.config_files, {})
        self.reporter.assert_reports(
            [fixture.error(reports.codes.INVALID_RESPONSE_FORMAT, node="NODE")]
        )

    def test_invalid_apiv1_data(self):
        requests = self.cmd.on_response(fixture_apiv1_response(data="invalid"))
        result = self.cmd.on_complete()

        self.assertEqual(requests, [])
        self.assertFalse(result.was_successful)
        self.assertEqual(result.config_files, {})
        self.reporter.assert_reports(
            [fixture.error(reports.codes.INVALID_RESPONSE_FORMAT, node="NODE")]
        )

    def test_incorrect_cluster_name(self):
        requests = self.cmd.on_response(
            fixture_apiv1_response(data=SyncConfigsDto("bad-name", {}))
        )
        result = self.cmd.on_complete()

        self.assertEqual(requests, [])
        self.assertFalse(result.was_successful)
        self.assertEqual(result.config_files, {})
        self.reporter.assert_reports(
            [
                fixture.error(
                    reports.codes.NODE_REPORTS_UNEXPECTED_CLUSTER_NAME,
                    cluster_name="test",
                    context=reports.dto.ReportItemContextDto(node="NODE"),
                )
            ]
        )

    def test_communication_error(self):
        requests = self.cmd.on_response(
            fixture_apiv1_response(response_code=500)
        )
        result = self.cmd.on_complete()

        self.assertEqual(requests, [])
        self.assertFalse(result.was_successful)
        self.assertEqual(result.config_files, {})
        self.reporter.assert_reports(
            [
                fixture.error(
                    reports.codes.NODE_COMMUNICATION_ERROR,
                    node="NODE",
                    command="api/v1/cfgsync-get-configs/v1",
                    reason="HTTP error: 500",
                    force_code=reports.codes.SKIP_OFFLINE_NODES,
                )
            ]
        )

    def test_reports_from_response_no_errors(self):
        requests = self.cmd.on_response(
            fixture_apiv1_response(
                report_list=[
                    reports.ReportItem.warning(
                        reports.messages.NodeReportsUnexpectedClusterName(
                            "WARN"
                        )
                    ).to_dto()
                ],
                data=SyncConfigsDto(
                    cluster_name="test",
                    configs={
                        file_type_codes.PCS_KNOWN_HOSTS: "foo",
                        file_type_codes.PCS_SETTINGS_CONF: "bar",
                    },
                ),
            )
        )
        result = self.cmd.on_complete()

        self.assertEqual(requests, [])
        self.assertFalse(result.was_successful)
        self.assertEqual(
            result.config_files,
            {
                file_type_codes.PCS_KNOWN_HOSTS: [ConfigInfo("NODE", "foo")],
                file_type_codes.PCS_SETTINGS_CONF: [ConfigInfo("NODE", "bar")],
            },
        )
        self.reporter.assert_reports(
            [
                fixture.warn(
                    reports.codes.NODE_REPORTS_UNEXPECTED_CLUSTER_NAME,
                    cluster_name="WARN",
                    context=reports.dto.ReportItemContextDto(node="NODE"),
                )
            ]
        )

    def test_reports_from_response_errors(self):
        requests = self.cmd.on_response(
            fixture_apiv1_response(
                report_list=[
                    reports.ReportItem.error(
                        reports.messages.NodeReportsUnexpectedClusterName(
                            "ERROR"
                        )
                    ).to_dto()
                ],
                data=SyncConfigsDto(
                    cluster_name="test",
                    configs={
                        file_type_codes.PCS_KNOWN_HOSTS: "foo",
                        file_type_codes.PCS_SETTINGS_CONF: "bar",
                    },
                ),
            )
        )
        result = self.cmd.on_complete()

        self.assertEqual(requests, [])
        self.assertFalse(result.was_successful)
        self.assertEqual(result.config_files, {})
        self.reporter.assert_reports(
            [
                fixture.error(
                    reports.codes.NODE_REPORTS_UNEXPECTED_CLUSTER_NAME,
                    cluster_name="ERROR",
                    context=reports.dto.ReportItemContextDto(node="NODE"),
                )
            ]
        )

    def test_apiv1_communication_error(self):
        requests = self.cmd.on_response(
            fixture_apiv1_response(
                com_status=COM_STATUS_ERROR, status_msg="Some error"
            )
        )
        result = self.cmd.on_complete()
        self.assertEqual(requests, [])
        self.assertFalse(result.was_successful)
        self.assertEqual(result.config_files, {})
        self.reporter.assert_reports(
            [
                fixture.error(
                    reports.codes.NODE_COMMUNICATION_COMMAND_UNSUCCESSFUL,
                    node="NODE",
                    command="api/v1/cfgsync-get-configs/v1",
                    reason="Some error",
                )
            ]
        )


class GetConfigsFallbackResponseProcessing(TestCase):
    def setUp(self):
        self.reporter = MockLibraryReportProcessor()
        self.cmd = GetConfigs(self.reporter, "test")

    def test_empty_response(self):
        requests = self.cmd.on_response(fixture_fallback_response())
        result = self.cmd.on_complete()

        self.assertEqual(requests, [])
        self.assertFalse(result.was_successful)
        self.assertEqual(result.config_files, {})
        self.reporter.assert_reports(
            [fixture.error(reports.codes.INVALID_RESPONSE_FORMAT, node="NODE")]
        )

    def test_single_response_with_files(self):
        requests = self.cmd.on_response(
            fixture_fallback_response(
                output=json.dumps(
                    {
                        "status": "ok",
                        "cluster_name": "test",
                        "configs": {
                            "known-hosts": {
                                "type": "file",
                                "text": "foo",
                            },
                            "pcs_settings.conf": {
                                "type": "file",
                                "text": "bar",
                            },
                        },
                    }
                )
            )
        )
        result = self.cmd.on_complete()

        self.assertEqual(requests, [])
        self.assertFalse(result.was_successful)
        self.assertEqual(
            result.config_files,
            {
                file_type_codes.PCS_KNOWN_HOSTS: [ConfigInfo("NODE", "foo")],
                file_type_codes.PCS_SETTINGS_CONF: [ConfigInfo("NODE", "bar")],
            },
        )
        self.reporter.assert_reports([])

    def test_multiple_responses_with_files(self):
        requests = self.cmd.on_response(
            fixture_fallback_response(
                node_label="NODE-1",
                output=json.dumps(
                    {
                        "status": "ok",
                        "cluster_name": "test",
                        "configs": {
                            "known-hosts": {
                                "type": "file",
                                "text": "foo-1",
                            },
                            "pcs_settings.conf": {
                                "type": "file",
                                "text": "bar-1",
                            },
                        },
                    }
                ),
            )
        )
        requests.extend(
            self.cmd.on_response(
                fixture_fallback_response(
                    node_label="NODE-2",
                    output=json.dumps(
                        {
                            "status": "ok",
                            "cluster_name": "test",
                            "configs": {
                                "known-hosts": {
                                    "type": "file",
                                    "text": "foo-2",
                                },
                                "pcs_settings.conf": {
                                    "type": "file",
                                    "text": "bar-2",
                                },
                            },
                        }
                    ),
                )
            )
        )
        result = self.cmd.on_complete()

        self.assertEqual(requests, [])
        self.assertTrue(result.was_successful)
        self.assertEqual(
            result.config_files,
            {
                file_type_codes.PCS_KNOWN_HOSTS: [
                    ConfigInfo("NODE-1", "foo-1"),
                    ConfigInfo("NODE-2", "foo-2"),
                ],
                file_type_codes.PCS_SETTINGS_CONF: [
                    ConfigInfo("NODE-1", "bar-1"),
                    ConfigInfo("NODE-2", "bar-2"),
                ],
            },
        )
        self.reporter.assert_reports([])

    def test_communication_error(self):
        requests = self.cmd.on_response(
            fixture_fallback_response(response_code=500)
        )
        result = self.cmd.on_complete()

        self.assertEqual(requests, [])
        self.assertFalse(result.was_successful)
        self.assertEqual(result.config_files, {})
        self.reporter.assert_reports(
            [
                fixture.error(
                    reports.codes.NODE_COMMUNICATION_ERROR,
                    node="NODE",
                    command="remote/get_configs",
                    reason="HTTP error: 500",
                    force_code=reports.codes.SKIP_OFFLINE_NODES,
                )
            ]
        )

    def test_404(self):
        requests = self.cmd.on_response(
            fixture_fallback_response(response_code=404)
        )
        result = self.cmd.on_complete()

        self.assertEqual(requests, [])
        self.assertFalse(result.was_successful)
        self.assertEqual(result.config_files, {})
        self.reporter.assert_reports(
            [
                fixture.error(
                    reports.codes.NODE_COMMUNICATION_ERROR_UNSUPPORTED_COMMAND,
                    node="NODE",
                    command="remote/get_configs",
                    reason="HTTP error: 404",
                    force_code=reports.codes.SKIP_OFFLINE_NODES,
                )
            ]
        )

    def test_not_in_cluster_status(self):
        requests = self.cmd.on_response(
            fixture_fallback_response(
                output=json.dumps({"status": "not_in_cluster"})
            )
        )
        result = self.cmd.on_complete()

        self.assertEqual(requests, [])
        self.assertFalse(result.was_successful)
        self.assertEqual(result.config_files, {})
        self.reporter.assert_reports(
            [
                fixture.error(
                    reports.codes.NODE_REPORTS_UNEXPECTED_CLUSTER_NAME,
                    cluster_name="test",
                    context=reports.dto.ReportItemContextDto(node="NODE"),
                )
            ]
        )

    def test_wrong_cluster_name_status(self):
        requests = self.cmd.on_response(
            fixture_fallback_response(
                output=json.dumps({"status": "wrong_cluster_name"})
            )
        )
        result = self.cmd.on_complete()

        self.assertEqual(requests, [])
        self.assertFalse(result.was_successful)
        self.assertEqual(result.config_files, {})
        self.reporter.assert_reports(
            [
                fixture.error(
                    reports.codes.NODE_REPORTS_UNEXPECTED_CLUSTER_NAME,
                    cluster_name="test",
                    context=reports.dto.ReportItemContextDto(node="NODE"),
                )
            ]
        )

    def test_wrong_cluster_name(self):
        requests = self.cmd.on_response(
            fixture_fallback_response(
                output=json.dumps({"status": "ok", "cluster_name": "bad-name"})
            )
        )
        result = self.cmd.on_complete()

        self.assertEqual(requests, [])
        self.assertFalse(result.was_successful)
        self.assertEqual(result.config_files, {})
        self.reporter.assert_reports(
            [
                fixture.error(
                    reports.codes.NODE_REPORTS_UNEXPECTED_CLUSTER_NAME,
                    cluster_name="test",
                    context=reports.dto.ReportItemContextDto(node="NODE"),
                )
            ]
        )

    def test_unknown_file(self):
        requests = self.cmd.on_response(
            fixture_fallback_response(
                output=json.dumps(
                    {
                        "status": "ok",
                        "cluster_name": "test",
                        "configs": {
                            "foobar": {"type": "file", "text": "foo"},
                        },
                    }
                )
            )
        )
        result = self.cmd.on_complete()

        self.assertEqual(requests, [])
        self.assertFalse(result.was_successful)
        self.assertEqual(result.config_files, {})
        self.reporter.assert_reports([])

    def test_not_a_file(self):
        requests = self.cmd.on_response(
            fixture_fallback_response(
                output=json.dumps(
                    {
                        "status": "ok",
                        "cluster_name": "test",
                        "configs": {
                            "known-hosts": {
                                "type": "not-a-file",
                                "text": "foo",
                            },
                        },
                    }
                )
            )
        )
        result = self.cmd.on_complete()

        self.assertEqual(requests, [])
        self.assertFalse(result.was_successful)
        self.assertEqual(result.config_files, {})
        self.reporter.assert_reports([])

    def test_invalid_data(self):
        requests = self.cmd.on_response(
            fixture_fallback_response(
                # the 'configs' key is missing
                output=json.dumps({"status": "ok", "cluster_name": "test"})
            )
        )
        result = self.cmd.on_complete()

        self.assertEqual(requests, [])
        self.assertFalse(result.was_successful)
        self.assertEqual(result.config_files, {})
        self.reporter.assert_reports(
            [fixture.error(reports.codes.INVALID_RESPONSE_FORMAT, node="NODE")]
        )


class GetConfigsResponseProcessing(TestCase):
    def setUp(self):
        self.reporter = MockLibraryReportProcessor()
        self.cmd = GetConfigs(self.reporter, "test")

    def test_multiple_responses_mixed_apiv1_fallback(self):
        requests = self.cmd.on_response(
            fixture_apiv1_response(
                node_label="NODE-1",
                data=SyncConfigsDto(
                    cluster_name="test",
                    configs={
                        file_type_codes.PCS_KNOWN_HOSTS: "foo-1",
                        file_type_codes.PCS_SETTINGS_CONF: "bar-1",
                    },
                ),
            )
        )
        requests.extend(
            self.cmd.on_response(
                fixture_fallback_response(
                    node_label="NODE-2",
                    output=json.dumps(
                        {
                            "status": "ok",
                            "cluster_name": "test",
                            "configs": {
                                "known-hosts": {
                                    "type": "file",
                                    "text": "foo-2",
                                },
                                "pcs_settings.conf": {
                                    "type": "file",
                                    "text": "bar-2",
                                },
                            },
                        }
                    ),
                )
            )
        )
        result = self.cmd.on_complete()

        self.assertEqual(requests, [])
        self.assertTrue(result.was_successful)
        self.assertEqual(
            result.config_files,
            {
                file_type_codes.PCS_KNOWN_HOSTS: [
                    ConfigInfo("NODE-1", "foo-1"),
                    ConfigInfo("NODE-2", "foo-2"),
                ],
                file_type_codes.PCS_SETTINGS_CONF: [
                    ConfigInfo("NODE-1", "bar-1"),
                    ConfigInfo("NODE-2", "bar-2"),
                ],
            },
        )
        self.reporter.assert_reports([])

    def test_multiple_responses_one_failed(self):
        requests = self.cmd.on_response(
            fixture_apiv1_response(
                node_label="NODE-1",
                data=SyncConfigsDto(
                    cluster_name="test",
                    configs={
                        file_type_codes.PCS_KNOWN_HOSTS: "foo-1",
                        file_type_codes.PCS_SETTINGS_CONF: "bar-1",
                    },
                ),
            )
        )
        requests.extend(
            self.cmd.on_response(
                fixture_apiv1_response(response_code=500, node_label="NODE-2")
            )
        )
        result = self.cmd.on_complete()

        self.assertEqual(requests, [])
        self.assertFalse(result.was_successful)
        self.assertEqual(
            result.config_files,
            {
                file_type_codes.PCS_KNOWN_HOSTS: [
                    ConfigInfo("NODE-1", "foo-1")
                ],
                file_type_codes.PCS_SETTINGS_CONF: [
                    ConfigInfo("NODE-1", "bar-1")
                ],
            },
        )
        self.reporter.assert_reports(
            [
                fixture.error(
                    reports.codes.NODE_COMMUNICATION_ERROR,
                    node="NODE-2",
                    command="api/v1/cfgsync-get-configs/v1",
                    reason="HTTP error: 500",
                    force_code=reports.codes.SKIP_OFFLINE_NODES,
                )
            ]
        )



from typing import Any

from pcs.cli.common.errors import CmdLineInputError
from pcs.cli.common.parse_args import Argv, InputModifiers


def description(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    """
    Options:
      * -f - CIB file
    """
    modifiers.ensure_only_supported("-f")

    if len(argv) == 1:
        element_id = argv[0]
        description = lib.cib.element_description_get(element_id)
        if description:
            print(description)
    elif len(argv) == 2:
        element_id, description = argv[0], argv[1]
        lib.cib.element_description_set(element_id, description)
    else:
        raise CmdLineInputError()



from typing import Optional, Sequence, cast

from pcs.common import reports
from pcs.common.file_type_codes import PCS_KNOWN_HOSTS, FileTypeCode
from pcs.common.host import PcsKnownHost
from pcs.common.node_communicator import Communicator, RequestTarget
from pcs.common.types import StringSequence
from pcs.lib.communication.pcs_cfgsync import SetConfigs, SetConfigsResult
from pcs.lib.communication.tools import run
from pcs.lib.file.toolbox import for_file_type as toolbox_for_file_type
from pcs.lib.host.config.facade import Facade as KnownHostsFacade
from pcs.lib.interface.config import SyncVersionFacadeInterface
from pcs.lib.pcs_cfgsync.fetcher import ConfigFetcher


def save_sync_new_version(
    file_type_code: FileTypeCode,
    file: SyncVersionFacadeInterface,
    cluster_name: str,
    target_list: Sequence[RequestTarget],
    node_communicator: Communicator,
    report_processor: reports.ReportProcessor,
    fetch_on_conflict: bool,
    reject_is_error: bool = True,
) -> tuple[bool, Optional[SyncVersionFacadeInterface]]:
    """
    Update the file data_version of the file and send it to targets. If any
    target contained newer version of the file and fetch_on_conflict is set
    to True, then fetch the newest version of the file from all of the targets.
    Returns True if conflict was detected, and newest file if it was fetched.

    file_type_code -- type of the file
    file -- the local file
    cluster_name -- name of the cluster
    target_list -- cluster nodes where we are sending the updated config
    node_communicator -- tool for communication with the nodes
    report_processor -- tool for reporting
    fetch_on_conflict -- if True, fetch the newest version of the file from
        the targets on conflict
    reject_is_error -- if True, rejected configs will be reported as errors
    """
    toolbox = toolbox_for_file_type(file_type_code)
    file.set_data_version(file.data_version + 1)
    com_cmd = SetConfigs(
        report_processor,
        cluster_name,
        {file_type_code: toolbox.exporter.export(file.config).decode("utf-8")},
        rejection_severity=reports.ReportItemSeverity.error()
        if reject_is_error
        else reports.ReportItemSeverity.info(),
    )
    com_cmd.set_targets(target_list)  # type: ignore[no-untyped-call]
    results = run(node_communicator, com_cmd)  # type: ignore[no-untyped-call]

    # we are only interested in the REJECTED state, so we can decide if we need
    # to fetch the newest version of the file from the cluster. The other
    # statuses/errors are already reported during the communication command
    # call
    if not any(
        results[node_name].get(file_type_code) == SetConfigsResult.REJECTED
        for node_name in results
    ):
        return False, None

    file_to_save = None
    if fetch_on_conflict:
        report_processor.report(
            reports.ReportItem.info(
                reports.messages.PcsCfgsyncFetchingNewestConfig(
                    [file_type_code], [target.label for target in target_list]
                )
            )
        )
        fetcher = ConfigFetcher(node_communicator, report_processor)
        newest_files, _ = fetcher.fetch(
            cluster_name, target_list, [file_type_code]
        )
        file_to_save = newest_files.get(file_type_code)
    return True, file_to_save


def save_sync_new_known_hosts(
    known_hosts_facade: KnownHostsFacade,
    new_hosts: Sequence[PcsKnownHost],
    hosts_to_remove: StringSequence,
    cluster_name: str,
    target_list: Sequence[RequestTarget],
    node_communicator: Communicator,
    report_processor: reports.ReportProcessor,
) -> tuple[bool, Optional[SyncVersionFacadeInterface]]:
    """
    Add and/or remove known hosts from the file, then send it to targets.
    If any target contained newer version of the file, then try to resolve the
    conflict by fetching the newest version of the file from all of the
    targets and merging the local file with the newest file from cluster. Then
    add and/or remove the new hosts to the merged file and send it to targets.
    If some node has newer version of the file even after this, then fetch
    the newest version. Returns True if there was conflict that the function
    could not resolve, and newest file if it was fetched.

    known_hosts_facade -- local known-hosts file
    new_hosts -- hosts to add
    hosts_to_remove -- hosts to remove
    cluster_name -- name of the cluster
    target_list -- cluster nodes where we are sending the updated config
    node_communicator -- tool for communication with the nodes
    report_processor -- tool for reporting
    """
    old_local_known_hosts = known_hosts_facade.config
    known_hosts_facade.remove_known_hosts(hosts_to_remove)
    known_hosts_facade.update_known_hosts(new_hosts)

    conflict_detected, new_file = save_sync_new_version(
        PCS_KNOWN_HOSTS,
        known_hosts_facade,
        cluster_name,
        target_list,
        node_communicator,
        report_processor,
        fetch_on_conflict=True,
        reject_is_error=False,
    )

    if not conflict_detected:
        return False, None

    if new_file is None:
        # we detected conflict, but got no new file to work with, so it does
        # not make sense to try sending the config again
        return True, None

    newest_known_hosts_from_cluster = cast(KnownHostsFacade, new_file)

    known_hosts_facade = KnownHostsFacade(old_local_known_hosts)
    known_hosts_facade.set_data_version(
        newest_known_hosts_from_cluster.data_version
    )
    # first add the tokens from the cluster then the new tokens, so we overwrite
    # any tokens that were previously in the configs with the new tokens
    known_hosts_facade.update_known_hosts(
        list(newest_known_hosts_from_cluster.known_hosts.values())
    )
    known_hosts_facade.remove_known_hosts(hosts_to_remove)
    known_hosts_facade.update_known_hosts(new_hosts)

    return save_sync_new_version(
        PCS_KNOWN_HOSTS,
        known_hosts_facade,
        cluster_name,
        target_list,
        node_communicator,
        report_processor,
        fetch_on_conflict=True,
    )



import json
from unittest import TestCase

from pcs.common.host import Destination, PcsKnownHost
from pcs.lib.file.json import JsonParserException
from pcs.lib.host.config.exporter import Exporter as KnownHostsExporter
from pcs.lib.host.config.parser import InvalidFileStructureException
from pcs.lib.host.config.parser import Parser as KnownHostsParser
from pcs.lib.host.config.types import KnownHosts

from pcs_test.tools.misc import read_test_resource

_FIXTURE_KNOWN_HOSTS_RAW_TEXT = read_test_resource("known-hosts")

_FIXTURE_KNOWN_HOSTS_STRUCTURE = KnownHosts(
    format_version=1,
    data_version=5,
    known_hosts={
        "node1": PcsKnownHost(
            name="node1",
            token="abcde",
            dest_list=[
                Destination("10.0.1.1", 2224),
                Destination("10.0.2.1", 2225),
            ],
        ),
        "node2": PcsKnownHost(
            name="node2",
            token="fghij",
            dest_list=[
                Destination("10.0.1.2", 2234),
                Destination("10.0.2.2", 2235),
            ],
        ),
    },
)


class Parser(TestCase):
    def test_invalid_json(self):
        with self.assertRaises(JsonParserException):
            KnownHostsParser.parse("<xml/>".encode())

    def test_root_not_object(self):
        with self.assertRaises(InvalidFileStructureException):
            KnownHostsParser.parse("[]")

    def test_invalid_file_content(self):
        with self.assertRaises(InvalidFileStructureException):
            KnownHostsParser.parse(json.dumps({"format_version": 1}))

    def test_valid_minimal(self):
        result = KnownHostsParser.parse(
            json.dumps(
                {"format_version": 1, "data_version": 10, "known_hosts": {}}
            )
        )
        self.assertEqual(
            KnownHosts(format_version=1, data_version=10, known_hosts=dict()),
            result,
        )

    def test_valid(self):
        result = KnownHostsParser.parse(_FIXTURE_KNOWN_HOSTS_RAW_TEXT)
        self.assertEqual(_FIXTURE_KNOWN_HOSTS_STRUCTURE, result)


class Exporter(TestCase):
    def test_ok(self):
        result = KnownHostsExporter.export(_FIXTURE_KNOWN_HOSTS_STRUCTURE)
        self.assertEqual(
            json.loads(_FIXTURE_KNOWN_HOSTS_RAW_TEXT), json.loads(result)
        )
